#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2009-2014, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Wrapper for gdi32.dll in ctypes.
"""

__revision__ = "$Id$"

from winappdbg.win32.defines import *
from winappdbg.win32.kernel32 import GetLastError, SetLastError

# ==============================================================================
# This is used later on to calculate the list of exported symbols.
_all = None
_all = set(vars().keys())
# ==============================================================================

# --- Helpers ------------------------------------------------------------------

# --- Types --------------------------------------------------------------------

# --- Constants ----------------------------------------------------------------

# GDI object types
OBJ_PEN = 1
OBJ_BRUSH = 2
OBJ_DC = 3
OBJ_METADC = 4
OBJ_PAL = 5
OBJ_FONT = 6
OBJ_BITMAP = 7
OBJ_REGION = 8
OBJ_METAFILE = 9
OBJ_MEMDC = 10
OBJ_EXTPEN = 11
OBJ_ENHMETADC = 12
OBJ_ENHMETAFILE = 13
OBJ_COLORSPACE = 14
GDI_OBJ_LAST = OBJ_COLORSPACE

# Ternary raster operations
SRCCOPY = 0x00CC0020  # dest = source
SRCPAINT = 0x00EE0086  # dest = source OR dest
SRCAND = 0x008800C6  # dest = source AND dest
SRCINVERT = 0x00660046  # dest = source XOR dest
SRCERASE = 0x00440328  # dest = source AND (NOT dest)
NOTSRCCOPY = 0x00330008  # dest = (NOT source)
NOTSRCERASE = 0x001100A6  # dest = (NOT src) AND (NOT dest)
MERGECOPY = 0x00C000CA  # dest = (source AND pattern)
MERGEPAINT = 0x00BB0226  # dest = (NOT source) OR dest
PATCOPY = 0x00F00021  # dest = pattern
PATPAINT = 0x00FB0A09  # dest = DPSnoo
PATINVERT = 0x005A0049  # dest = pattern XOR dest
DSTINVERT = 0x00550009  # dest = (NOT dest)
BLACKNESS = 0x00000042  # dest = BLACK
WHITENESS = 0x00FF0062  # dest = WHITE
NOMIRRORBITMAP = 0x80000000  # Do not Mirror the bitmap in this call
CAPTUREBLT = 0x40000000  # Include layered windows

# Region flags
ERROR = 0
NULLREGION = 1
SIMPLEREGION = 2
COMPLEXREGION = 3
RGN_ERROR = ERROR

# CombineRgn() styles
RGN_AND = 1
RGN_OR = 2
RGN_XOR = 3
RGN_DIFF = 4
RGN_COPY = 5
RGN_MIN = RGN_AND
RGN_MAX = RGN_COPY

# StretchBlt() modes
BLACKONWHITE = 1
WHITEONBLACK = 2
COLORONCOLOR = 3
HALFTONE = 4
MAXSTRETCHBLTMODE = 4
STRETCH_ANDSCANS = BLACKONWHITE
STRETCH_ORSCANS = WHITEONBLACK
STRETCH_DELETESCANS = COLORONCOLOR
STRETCH_HALFTONE = HALFTONE

# PolyFill() modes
ALTERNATE = 1
WINDING = 2
POLYFILL_LAST = 2

# Layout orientation options
LAYOUT_RTL = 0x00000001  # Right to left
LAYOUT_BTT = 0x00000002  # Bottom to top
LAYOUT_VBH = 0x00000004  # Vertical before horizontal
LAYOUT_ORIENTATIONMASK = LAYOUT_RTL + LAYOUT_BTT + LAYOUT_VBH
LAYOUT_BITMAPORIENTATIONPRESERVED = 0x00000008

# Stock objects
WHITE_BRUSH = 0
LTGRAY_BRUSH = 1
GRAY_BRUSH = 2
DKGRAY_BRUSH = 3
BLACK_BRUSH = 4
NULL_BRUSH = 5
HOLLOW_BRUSH = NULL_BRUSH
WHITE_PEN = 6
BLACK_PEN = 7
NULL_PEN = 8
OEM_FIXED_FONT = 10
ANSI_FIXED_FONT = 11
ANSI_VAR_FONT = 12
SYSTEM_FONT = 13
DEVICE_DEFAULT_FONT = 14
DEFAULT_PALETTE = 15
SYSTEM_FIXED_FONT = 16

# Metafile functions
META_SETBKCOLOR = 0x0201
META_SETBKMODE = 0x0102
META_SETMAPMODE = 0x0103
META_SETROP2 = 0x0104
META_SETRELABS = 0x0105
META_SETPOLYFILLMODE = 0x0106
META_SETSTRETCHBLTMODE = 0x0107
META_SETTEXTCHAREXTRA = 0x0108
META_SETTEXTCOLOR = 0x0209
META_SETTEXTJUSTIFICATION = 0x020A
META_SETWINDOWORG = 0x020B
META_SETWINDOWEXT = 0x020C
META_SETVIEWPORTORG = 0x020D
META_SETVIEWPORTEXT = 0x020E
META_OFFSETWINDOWORG = 0x020F
META_SCALEWINDOWEXT = 0x0410
META_OFFSETVIEWPORTORG = 0x0211
META_SCALEVIEWPORTEXT = 0x0412
META_LINETO = 0x0213
META_MOVETO = 0x0214
META_EXCLUDECLIPRECT = 0x0415
META_INTERSECTCLIPRECT = 0x0416
META_ARC = 0x0817
META_ELLIPSE = 0x0418
META_FLOODFILL = 0x0419
META_PIE = 0x081A
META_RECTANGLE = 0x041B
META_ROUNDRECT = 0x061C
META_PATBLT = 0x061D
META_SAVEDC = 0x001E
META_SETPIXEL = 0x041F
META_OFFSETCLIPRGN = 0x0220
META_TEXTOUT = 0x0521
META_BITBLT = 0x0922
META_STRETCHBLT = 0x0B23
META_POLYGON = 0x0324
META_POLYLINE = 0x0325
META_ESCAPE = 0x0626
META_RESTOREDC = 0x0127
META_FILLREGION = 0x0228
META_FRAMEREGION = 0x0429
META_INVERTREGION = 0x012A
META_PAINTREGION = 0x012B
META_SELECTCLIPREGION = 0x012C
META_SELECTOBJECT = 0x012D
META_SETTEXTALIGN = 0x012E
META_CHORD = 0x0830
META_SETMAPPERFLAGS = 0x0231
META_EXTTEXTOUT = 0x0A32
META_SETDIBTODEV = 0x0D33
META_SELECTPALETTE = 0x0234
META_REALIZEPALETTE = 0x0035
META_ANIMATEPALETTE = 0x0436
META_SETPALENTRIES = 0x0037
META_POLYPOLYGON = 0x0538
META_RESIZEPALETTE = 0x0139
META_DIBBITBLT = 0x0940
META_DIBSTRETCHBLT = 0x0B41
META_DIBCREATEPATTERNBRUSH = 0x0142
META_STRETCHDIB = 0x0F43
META_EXTFLOODFILL = 0x0548
META_SETLAYOUT = 0x0149
META_DELETEOBJECT = 0x01F0
META_CREATEPALETTE = 0x00F7
META_CREATEPATTERNBRUSH = 0x01F9
META_CREATEPENINDIRECT = 0x02FA
META_CREATEFONTINDIRECT = 0x02FB
META_CREATEBRUSHINDIRECT = 0x02FC
META_CREATEREGION = 0x06FF

# Metafile escape codes
NEWFRAME = 1
ABORTDOC = 2
NEXTBAND = 3
SETCOLORTABLE = 4
GETCOLORTABLE = 5
FLUSHOUTPUT = 6
DRAFTMODE = 7
QUERYESCSUPPORT = 8
SETABORTPROC = 9
STARTDOC = 10
ENDDOC = 11
GETPHYSPAGESIZE = 12
GETPRINTINGOFFSET = 13
GETSCALINGFACTOR = 14
MFCOMMENT = 15
GETPENWIDTH = 16
SETCOPYCOUNT = 17
SELECTPAPERSOURCE = 18
DEVICEDATA = 19
PASSTHROUGH = 19
GETTECHNOLGY = 20
GETTECHNOLOGY = 20
SETLINECAP = 21
SETLINEJOIN = 22
SETMITERLIMIT = 23
BANDINFO = 24
DRAWPATTERNRECT = 25
GETVECTORPENSIZE = 26
GETVECTORBRUSHSIZE = 27
ENABLEDUPLEX = 28
GETSETPAPERBINS = 29
GETSETPRINTORIENT = 30
ENUMPAPERBINS = 31
SETDIBSCALING = 32
EPSPRINTING = 33
ENUMPAPERMETRICS = 34
GETSETPAPERMETRICS = 35
POSTSCRIPT_DATA = 37
POSTSCRIPT_IGNORE = 38
MOUSETRAILS = 39
GETDEVICEUNITS = 42
GETEXTENDEDTEXTMETRICS = 256
GETEXTENTTABLE = 257
GETPAIRKERNTABLE = 258
GETTRACKKERNTABLE = 259
EXTTEXTOUT = 512
GETFACENAME = 513
DOWNLOADFACE = 514
ENABLERELATIVEWIDTHS = 768
ENABLEPAIRKERNING = 769
SETKERNTRACK = 770
SETALLJUSTVALUES = 771
SETCHARSET = 772
STRETCHBLT = 2048
METAFILE_DRIVER = 2049
GETSETSCREENPARAMS = 3072
QUERYDIBSUPPORT = 3073
BEGIN_PATH = 4096
CLIP_TO_PATH = 4097
END_PATH = 4098
EXT_DEVICE_CAPS = 4099
RESTORE_CTM = 4100
SAVE_CTM = 4101
SET_ARC_DIRECTION = 4102
SET_BACKGROUND_COLOR = 4103
SET_POLY_MODE = 4104
SET_SCREEN_ANGLE = 4105
SET_SPREAD = 4106
TRANSFORM_CTM = 4107
SET_CLIP_BOX = 4108
SET_BOUNDS = 4109
SET_MIRROR_MODE = 4110
OPENCHANNEL = 4110
DOWNLOADHEADER = 4111
CLOSECHANNEL = 4112
POSTSCRIPT_PASSTHROUGH = 4115
ENCAPSULATED_POSTSCRIPT = 4116
POSTSCRIPT_IDENTIFY = 4117
POSTSCRIPT_INJECTION = 4118
CHECKJPEGFORMAT = 4119
CHECKPNGFORMAT = 4120
GET_PS_FEATURESETTING = 4121
GDIPLUS_TS_QUERYVER = 4122
GDIPLUS_TS_RECORD = 4123
SPCLPASSTHROUGH2 = 4568

# --- Structures ---------------------------------------------------------------


# typedef struct _RECT {
#   LONG left;
#   LONG top;
#   LONG right;
#   LONG bottom;
# }RECT, *PRECT;
class RECT(Structure):
    _fields_ = [
        ("left", LONG),
        ("top", LONG),
        ("right", LONG),
        ("bottom", LONG),
    ]


PRECT = POINTER(RECT)
LPRECT = PRECT


# typedef struct tagPOINT {
#   LONG x;
#   LONG y;
# } POINT;
class POINT(Structure):
    _fields_ = [
        ("x", LONG),
        ("y", LONG),
    ]


PPOINT = POINTER(POINT)
LPPOINT = PPOINT


# typedef struct tagBITMAP {
#   LONG   bmType;
#   LONG   bmWidth;
#   LONG   bmHeight;
#   LONG   bmWidthBytes;
#   WORD   bmPlanes;
#   WORD   bmBitsPixel;
#   LPVOID bmBits;
# } BITMAP, *PBITMAP;
class BITMAP(Structure):
    _fields_ = [
        ("bmType", LONG),
        ("bmWidth", LONG),
        ("bmHeight", LONG),
        ("bmWidthBytes", LONG),
        ("bmPlanes", WORD),
        ("bmBitsPixel", WORD),
        ("bmBits", LPVOID),
    ]


PBITMAP = POINTER(BITMAP)
LPBITMAP = PBITMAP

# --- High level classes -------------------------------------------------------

# --- gdi32.dll ----------------------------------------------------------------


# HDC GetDC(
#   __in  HWND hWnd
# );
def GetDC(hWnd):
    _GetDC = windll.gdi32.GetDC
    _GetDC.argtypes = [HWND]
    _GetDC.restype = HDC
    _GetDC.errcheck = RaiseIfZero
    return _GetDC(hWnd)


# HDC GetWindowDC(
#   __in  HWND hWnd
# );
def GetWindowDC(hWnd):
    _GetWindowDC = windll.gdi32.GetWindowDC
    _GetWindowDC.argtypes = [HWND]
    _GetWindowDC.restype = HDC
    _GetWindowDC.errcheck = RaiseIfZero
    return _GetWindowDC(hWnd)


# int ReleaseDC(
#   __in  HWND hWnd,
#   __in  HDC hDC
# );
def ReleaseDC(hWnd, hDC):
    _ReleaseDC = windll.gdi32.ReleaseDC
    _ReleaseDC.argtypes = [HWND, HDC]
    _ReleaseDC.restype = ctypes.c_int
    _ReleaseDC.errcheck = RaiseIfZero
    _ReleaseDC(hWnd, hDC)


# HGDIOBJ SelectObject(
#   __in  HDC hdc,
#   __in  HGDIOBJ hgdiobj
# );
def SelectObject(hdc, hgdiobj):
    _SelectObject = windll.gdi32.SelectObject
    _SelectObject.argtypes = [HDC, HGDIOBJ]
    _SelectObject.restype = HGDIOBJ
    _SelectObject.errcheck = RaiseIfZero
    return _SelectObject(hdc, hgdiobj)


# HGDIOBJ GetStockObject(
#   __in  int fnObject
# );
def GetStockObject(fnObject):
    _GetStockObject = windll.gdi32.GetStockObject
    _GetStockObject.argtypes = [ctypes.c_int]
    _GetStockObject.restype = HGDIOBJ
    _GetStockObject.errcheck = RaiseIfZero
    return _GetStockObject(fnObject)


# DWORD GetObjectType(
#   __in  HGDIOBJ h
# );
def GetObjectType(h):
    _GetObjectType = windll.gdi32.GetObjectType
    _GetObjectType.argtypes = [HGDIOBJ]
    _GetObjectType.restype = DWORD
    _GetObjectType.errcheck = RaiseIfZero
    return _GetObjectType(h)


# int GetObject(
#   __in   HGDIOBJ hgdiobj,
#   __in   int cbBuffer,
#   __out  LPVOID lpvObject
# );
def GetObject(hgdiobj, cbBuffer=None, lpvObject=None):
    _GetObject = windll.gdi32.GetObject
    _GetObject.argtypes = [HGDIOBJ, ctypes.c_int, LPVOID]
    _GetObject.restype = ctypes.c_int
    _GetObject.errcheck = RaiseIfZero

    # Both cbBuffer and lpvObject can be omitted, the correct
    # size and structure to return are automatically deduced.
    # If lpvObject is given it must be a ctypes object, not a pointer.
    # Always returns a ctypes object.

    if cbBuffer is not None:
        if lpvObject is None:
            lpvObject = ctypes.create_string_buffer("", cbBuffer)
    elif lpvObject is not None:
        cbBuffer = sizeof(lpvObject)
    else:  # most likely case, both are None
        t = GetObjectType(hgdiobj)
        if t == OBJ_PEN:
            cbBuffer = sizeof(LOGPEN)
            lpvObject = LOGPEN()
        elif t == OBJ_BRUSH:
            cbBuffer = sizeof(LOGBRUSH)
            lpvObject = LOGBRUSH()
        elif t == OBJ_PAL:
            cbBuffer = _GetObject(hgdiobj, 0, None)
            lpvObject = (WORD * (cbBuffer // sizeof(WORD)))()
        elif t == OBJ_FONT:
            cbBuffer = sizeof(LOGFONT)
            lpvObject = LOGFONT()
        elif t == OBJ_BITMAP:  # try the two possible types of bitmap
            cbBuffer = sizeof(DIBSECTION)
            lpvObject = DIBSECTION()
            try:
                _GetObject(hgdiobj, cbBuffer, byref(lpvObject))
                return lpvObject
            except WindowsError:
                cbBuffer = sizeof(BITMAP)
                lpvObject = BITMAP()
        elif t == OBJ_EXTPEN:
            cbBuffer = sizeof(LOGEXTPEN)
            lpvObject = LOGEXTPEN()
        else:
            cbBuffer = _GetObject(hgdiobj, 0, None)
            lpvObject = ctypes.create_string_buffer("", cbBuffer)
    _GetObject(hgdiobj, cbBuffer, byref(lpvObject))
    return lpvObject


# LONG GetBitmapBits(
#   __in   HBITMAP hbmp,
#   __in   LONG cbBuffer,
#   __out  LPVOID lpvBits
# );
def GetBitmapBits(hbmp):
    _GetBitmapBits = windll.gdi32.GetBitmapBits
    _GetBitmapBits.argtypes = [HBITMAP, LONG, LPVOID]
    _GetBitmapBits.restype = LONG
    _GetBitmapBits.errcheck = RaiseIfZero

    bitmap = GetObject(hbmp, lpvObject=BITMAP())
    cbBuffer = bitmap.bmWidthBytes * bitmap.bmHeight
    lpvBits = ctypes.create_string_buffer("", cbBuffer)
    _GetBitmapBits(hbmp, cbBuffer, byref(lpvBits))
    return lpvBits.raw


# HBITMAP CreateBitmapIndirect(
#   __in  const BITMAP *lpbm
# );
def CreateBitmapIndirect(lpbm):
    _CreateBitmapIndirect = windll.gdi32.CreateBitmapIndirect
    _CreateBitmapIndirect.argtypes = [PBITMAP]
    _CreateBitmapIndirect.restype = HBITMAP
    _CreateBitmapIndirect.errcheck = RaiseIfZero
    return _CreateBitmapIndirect(lpbm)


# ==============================================================================
# This calculates the list of exported symbols.
_all = set(vars().keys()).difference(_all)
__all__ = [_x for _x in _all if not _x.startswith("_")]
__all__.sort()
# ==============================================================================



import pytest
from tests_python.debugger_unittest import IS_PY36_OR_GREATER, IS_CPYTHON
from tests_python.debug_constants import TEST_CYTHON

pytestmark = pytest.mark.skipif(not IS_PY36_OR_GREATER or not IS_CPYTHON or not TEST_CYTHON, reason="Requires CPython >= 3.6")
import sys
import unittest
from _pydevd_frame_eval.vendored.bytecode import Label, Instr, Compare, Bytecode, ControlFlowGraph
from _pydevd_frame_eval.vendored.bytecode import peephole_opt
from _pydevd_frame_eval.vendored.bytecode.tests import TestCase, dump_bytecode
from unittest import mock


class Tests(TestCase):
    maxDiff = 80 * 100

    def optimize_blocks(self, code):
        if isinstance(code, Bytecode):
            code = ControlFlowGraph.from_bytecode(code)
        optimizer = peephole_opt.PeepholeOptimizer()
        optimizer.optimize_cfg(code)
        return code

    def check(self, code, *expected):
        if isinstance(code, Bytecode):
            code = ControlFlowGraph.from_bytecode(code)
        optimizer = peephole_opt.PeepholeOptimizer()
        optimizer.optimize_cfg(code)
        code = code.to_bytecode()

        try:
            self.assertEqual(code, expected)
        except AssertionError:
            print("Optimized code:")
            dump_bytecode(code)

            print("Expected code:")
            for instr in expected:
                print(instr)

            raise

    def check_dont_optimize(self, code):
        code = ControlFlowGraph.from_bytecode(code)
        noopt = code.to_bytecode()

        optim = self.optimize_blocks(code)
        optim = optim.to_bytecode()
        self.assertEqual(optim, noopt)

    def test_unary_op(self):
        def check_unary_op(op, value, result):
            code = Bytecode([Instr("LOAD_CONST", value), Instr(op), Instr("STORE_NAME", "x")])
            self.check(code, Instr("LOAD_CONST", result), Instr("STORE_NAME", "x"))

        check_unary_op("UNARY_POSITIVE", 2, 2)
        check_unary_op("UNARY_NEGATIVE", 3, -3)
        check_unary_op("UNARY_INVERT", 5, -6)

    def test_binary_op(self):
        def check_bin_op(left, op, right, result):
            code = Bytecode(
                [
                    Instr("LOAD_CONST", left),
                    Instr("LOAD_CONST", right),
                    Instr(op),
                    Instr("STORE_NAME", "x"),
                ]
            )
            self.check(code, Instr("LOAD_CONST", result), Instr("STORE_NAME", "x"))

        check_bin_op(10, "BINARY_ADD", 20, 30)
        check_bin_op(5, "BINARY_SUBTRACT", 1, 4)
        check_bin_op(5, "BINARY_MULTIPLY", 3, 15)
        check_bin_op(10, "BINARY_TRUE_DIVIDE", 3, 10 / 3)
        check_bin_op(10, "BINARY_FLOOR_DIVIDE", 3, 3)
        check_bin_op(10, "BINARY_MODULO", 3, 1)
        check_bin_op(2, "BINARY_POWER", 8, 256)
        check_bin_op(1, "BINARY_LSHIFT", 3, 8)
        check_bin_op(16, "BINARY_RSHIFT", 3, 2)
        check_bin_op(10, "BINARY_AND", 3, 2)
        check_bin_op(2, "BINARY_OR", 3, 3)
        check_bin_op(2, "BINARY_XOR", 3, 1)

    def test_combined_unary_bin_ops(self):
        # x = 1 + 3 + 7
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 3),
                Instr("BINARY_ADD"),
                Instr("LOAD_CONST", 7),
                Instr("BINARY_ADD"),
                Instr("STORE_NAME", "x"),
            ]
        )
        self.check(code, Instr("LOAD_CONST", 11), Instr("STORE_NAME", "x"))

        # x = ~(~(5))
        code = Bytecode(
            [
                Instr("LOAD_CONST", 5),
                Instr("UNARY_INVERT"),
                Instr("UNARY_INVERT"),
                Instr("STORE_NAME", "x"),
            ]
        )
        self.check(code, Instr("LOAD_CONST", 5), Instr("STORE_NAME", "x"))

        # "events = [(0, 'call'), (1, 'line'), (-(3), 'call')]"
        code = Bytecode(
            [
                Instr("LOAD_CONST", 0),
                Instr("LOAD_CONST", "call"),
                Instr("BUILD_TUPLE", 2),
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", "line"),
                Instr("BUILD_TUPLE", 2),
                Instr("LOAD_CONST", 3),
                Instr("UNARY_NEGATIVE"),
                Instr("LOAD_CONST", "call"),
                Instr("BUILD_TUPLE", 2),
                Instr("BUILD_LIST", 3),
                Instr("STORE_NAME", "events"),
            ]
        )
        self.check(
            code,
            Instr("LOAD_CONST", (0, "call")),
            Instr("LOAD_CONST", (1, "line")),
            Instr("LOAD_CONST", (-3, "call")),
            Instr("BUILD_LIST", 3),
            Instr("STORE_NAME", "events"),
        )

        # 'x = (1,) + (0,) * 8'
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("BUILD_TUPLE", 1),
                Instr("LOAD_CONST", 0),
                Instr("BUILD_TUPLE", 1),
                Instr("LOAD_CONST", 8),
                Instr("BINARY_MULTIPLY"),
                Instr("BINARY_ADD"),
                Instr("STORE_NAME", "x"),
            ]
        )
        zeros = (0,) * 8
        result = (1,) + zeros
        self.check(code, Instr("LOAD_CONST", result), Instr("STORE_NAME", "x"))

    def test_max_size(self):
        max_size = 3
        with mock.patch.object(peephole_opt, "MAX_SIZE", max_size):
            # optimized binary operation: size <= maximum size
            #
            # (9,) * size
            size = max_size
            result = (9,) * size
            code = Bytecode(
                [
                    Instr("LOAD_CONST", 9),
                    Instr("BUILD_TUPLE", 1),
                    Instr("LOAD_CONST", size),
                    Instr("BINARY_MULTIPLY"),
                    Instr("STORE_NAME", "x"),
                ]
            )
            self.check(code, Instr("LOAD_CONST", result), Instr("STORE_NAME", "x"))

            # don't optimize  binary operation: size > maximum size
            #
            # x = (9,) * size
            size = max_size + 1
            code = Bytecode(
                [
                    Instr("LOAD_CONST", 9),
                    Instr("BUILD_TUPLE", 1),
                    Instr("LOAD_CONST", size),
                    Instr("BINARY_MULTIPLY"),
                    Instr("STORE_NAME", "x"),
                ]
            )
            self.check(
                code,
                Instr("LOAD_CONST", (9,)),
                Instr("LOAD_CONST", size),
                Instr("BINARY_MULTIPLY"),
                Instr("STORE_NAME", "x"),
            )

    def test_bin_op_dont_optimize(self):
        # 1 / 0
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 0),
                Instr("BINARY_TRUE_DIVIDE"),
                Instr("POP_TOP"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        self.check_dont_optimize(code)

        # 1 // 0
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 0),
                Instr("BINARY_FLOOR_DIVIDE"),
                Instr("POP_TOP"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        self.check_dont_optimize(code)

        # 1 % 0
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 0),
                Instr("BINARY_MODULO"),
                Instr("POP_TOP"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        self.check_dont_optimize(code)

        # 1 % 1j
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 1j),
                Instr("BINARY_MODULO"),
                Instr("POP_TOP"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        self.check_dont_optimize(code)

    def test_build_tuple(self):
        # x = (1, 2, 3)
        code = Bytecode(
            [
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 2),
                Instr("LOAD_CONST", 3),
                Instr("BUILD_TUPLE", 3),
                Instr("STORE_NAME", "x"),
            ]
        )
        self.check(code, Instr("LOAD_CONST", (1, 2, 3)), Instr("STORE_NAME", "x"))

    def test_build_list(self):
        # test = x in [1, 2, 3]
        code = Bytecode(
            [
                Instr("LOAD_NAME", "x"),
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 2),
                Instr("LOAD_CONST", 3),
                Instr("BUILD_LIST", 3),
                Instr("COMPARE_OP", Compare.IN),
                Instr("STORE_NAME", "test"),
            ]
        )

        self.check(
            code,
            Instr("LOAD_NAME", "x"),
            Instr("LOAD_CONST", (1, 2, 3)),
            Instr("COMPARE_OP", Compare.IN),
            Instr("STORE_NAME", "test"),
        )

    def test_build_list_unpack_seq(self):
        for build_list in ("BUILD_TUPLE", "BUILD_LIST"):
            # x, = [a]
            code = Bytecode(
                [
                    Instr("LOAD_NAME", "a"),
                    Instr(build_list, 1),
                    Instr("UNPACK_SEQUENCE", 1),
                    Instr("STORE_NAME", "x"),
                ]
            )
            self.check(code, Instr("LOAD_NAME", "a"), Instr("STORE_NAME", "x"))

            # x, y = [a, b]
            code = Bytecode(
                [
                    Instr("LOAD_NAME", "a"),
                    Instr("LOAD_NAME", "b"),
                    Instr(build_list, 2),
                    Instr("UNPACK_SEQUENCE", 2),
                    Instr("STORE_NAME", "x"),
                    Instr("STORE_NAME", "y"),
                ]
            )
            self.check(
                code,
                Instr("LOAD_NAME", "a"),
                Instr("LOAD_NAME", "b"),
                Instr("ROT_TWO"),
                Instr("STORE_NAME", "x"),
                Instr("STORE_NAME", "y"),
            )

            # x, y, z = [a, b, c]
            code = Bytecode(
                [
                    Instr("LOAD_NAME", "a"),
                    Instr("LOAD_NAME", "b"),
                    Instr("LOAD_NAME", "c"),
                    Instr(build_list, 3),
                    Instr("UNPACK_SEQUENCE", 3),
                    Instr("STORE_NAME", "x"),
                    Instr("STORE_NAME", "y"),
                    Instr("STORE_NAME", "z"),
                ]
            )
            self.check(
                code,
                Instr("LOAD_NAME", "a"),
                Instr("LOAD_NAME", "b"),
                Instr("LOAD_NAME", "c"),
                Instr("ROT_THREE"),
                Instr("ROT_TWO"),
                Instr("STORE_NAME", "x"),
                Instr("STORE_NAME", "y"),
                Instr("STORE_NAME", "z"),
            )

    def test_build_tuple_unpack_seq_const(self):
        # x, y = (3, 4)
        code = Bytecode(
            [
                Instr("LOAD_CONST", 3),
                Instr("LOAD_CONST", 4),
                Instr("BUILD_TUPLE", 2),
                Instr("UNPACK_SEQUENCE", 2),
                Instr("STORE_NAME", "x"),
                Instr("STORE_NAME", "y"),
            ]
        )
        self.check(
            code,
            Instr("LOAD_CONST", (3, 4)),
            Instr("UNPACK_SEQUENCE", 2),
            Instr("STORE_NAME", "x"),
            Instr("STORE_NAME", "y"),
        )

    def test_build_list_unpack_seq_const(self):
        # x, y, z = [3, 4, 5]
        code = Bytecode(
            [
                Instr("LOAD_CONST", 3),
                Instr("LOAD_CONST", 4),
                Instr("LOAD_CONST", 5),
                Instr("BUILD_LIST", 3),
                Instr("UNPACK_SEQUENCE", 3),
                Instr("STORE_NAME", "x"),
                Instr("STORE_NAME", "y"),
                Instr("STORE_NAME", "z"),
            ]
        )
        self.check(
            code,
            Instr("LOAD_CONST", 5),
            Instr("LOAD_CONST", 4),
            Instr("LOAD_CONST", 3),
            Instr("STORE_NAME", "x"),
            Instr("STORE_NAME", "y"),
            Instr("STORE_NAME", "z"),
        )

    def test_build_set(self):
        # test = x in {1, 2, 3}
        code = Bytecode(
            [
                Instr("LOAD_NAME", "x"),
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", 2),
                Instr("LOAD_CONST", 3),
                Instr("BUILD_SET", 3),
                Instr("COMPARE_OP", Compare.IN),
                Instr("STORE_NAME", "test"),
            ]
        )

        self.check(
            code,
            Instr("LOAD_NAME", "x"),
            Instr("LOAD_CONST", frozenset((1, 2, 3))),
            Instr("COMPARE_OP", Compare.IN),
            Instr("STORE_NAME", "test"),
        )

    def test_compare_op_unary_not(self):
        for op, not_op in (
            (Compare.IN, Compare.NOT_IN),  # in => not in
            (Compare.NOT_IN, Compare.IN),  # not in => in
            (Compare.IS, Compare.IS_NOT),  # is => is not
            (Compare.IS_NOT, Compare.IS),  # is not => is
        ):
            code = Bytecode(
                [
                    Instr("LOAD_NAME", "a"),
                    Instr("LOAD_NAME", "b"),
                    Instr("COMPARE_OP", op),
                    Instr("UNARY_NOT"),
                    Instr("STORE_NAME", "x"),
                ]
            )
            self.check(
                code,
                Instr("LOAD_NAME", "a"),
                Instr("LOAD_NAME", "b"),
                Instr("COMPARE_OP", not_op),
                Instr("STORE_NAME", "x"),
            )

        # don't optimize:
        # x = not (a and b is True)
        label_instr5 = Label()
        code = Bytecode(
            [
                Instr("LOAD_NAME", "a"),
                Instr("JUMP_IF_FALSE_OR_POP", label_instr5),
                Instr("LOAD_NAME", "b"),
                Instr("LOAD_CONST", True),
                Instr("COMPARE_OP", Compare.IS),
                label_instr5,
                Instr("UNARY_NOT"),
                Instr("STORE_NAME", "x"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        self.check_dont_optimize(code)

    def test_dont_optimize(self):
        # x = 3 < 5
        code = Bytecode(
            [
                Instr("LOAD_CONST", 3),
                Instr("LOAD_CONST", 5),
                Instr("COMPARE_OP", Compare.LT),
                Instr("STORE_NAME", "x"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        self.check_dont_optimize(code)

        # x = (10, 20, 30)[1:]
        code = Bytecode(
            [
                Instr("LOAD_CONST", (10, 20, 30)),
                Instr("LOAD_CONST", 1),
                Instr("LOAD_CONST", None),
                Instr("BUILD_SLICE", 2),
                Instr("BINARY_SUBSCR"),
                Instr("STORE_NAME", "x"),
            ]
        )
        self.check_dont_optimize(code)

    def test_optimize_code_obj(self):
        # Test optimize() method with a code object
        #
        # x = 3 + 5 => x = 8
        noopt = Bytecode(
            [
                Instr("LOAD_CONST", 3),
                Instr("LOAD_CONST", 5),
                Instr("BINARY_ADD"),
                Instr("STORE_NAME", "x"),
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )
        noopt = noopt.to_code()

        optimizer = peephole_opt.PeepholeOptimizer()
        optim = optimizer.optimize(noopt)

        code = Bytecode.from_code(optim)
        self.assertEqual(
            code,
            [
                Instr("LOAD_CONST", 8, lineno=1),
                Instr("STORE_NAME", "x", lineno=1),
                Instr("LOAD_CONST", None, lineno=1),
                Instr("RETURN_VALUE", lineno=1),
            ],
        )

    def test_return_value(self):
        # return+return: remove second return
        #
        #     def func():
        #         return 4
        #         return 5
        code = Bytecode(
            [
                Instr("LOAD_CONST", 4, lineno=2),
                Instr("RETURN_VALUE", lineno=2),
                Instr("LOAD_CONST", 5, lineno=3),
                Instr("RETURN_VALUE", lineno=3),
            ]
        )
        code = ControlFlowGraph.from_bytecode(code)
        self.check(code, Instr("LOAD_CONST", 4, lineno=2), Instr("RETURN_VALUE", lineno=2))

        # return+return + return+return: remove second and fourth return
        #
        #     def func():
        #         return 4
        #         return 5
        #         return 6
        #         return 7
        code = Bytecode(
            [
                Instr("LOAD_CONST", 4, lineno=2),
                Instr("RETURN_VALUE", lineno=2),
                Instr("LOAD_CONST", 5, lineno=3),
                Instr("RETURN_VALUE", lineno=3),
                Instr("LOAD_CONST", 6, lineno=4),
                Instr("RETURN_VALUE", lineno=4),
                Instr("LOAD_CONST", 7, lineno=5),
                Instr("RETURN_VALUE", lineno=5),
            ]
        )
        code = ControlFlowGraph.from_bytecode(code)
        self.check(code, Instr("LOAD_CONST", 4, lineno=2), Instr("RETURN_VALUE", lineno=2))

        # return + JUMP_ABSOLUTE: remove JUMP_ABSOLUTE
        # while 1:
        #     return 7
        if sys.version_info < (3, 8):
            setup_loop = Label()
            return_label = Label()
            code = Bytecode(
                [
                    setup_loop,
                    Instr("SETUP_LOOP", return_label, lineno=2),
                    Instr("LOAD_CONST", 7, lineno=3),
                    Instr("RETURN_VALUE", lineno=3),
                    Instr("JUMP_ABSOLUTE", setup_loop, lineno=3),
                    Instr("POP_BLOCK", lineno=3),
                    return_label,
                    Instr("LOAD_CONST", None, lineno=3),
                    Instr("RETURN_VALUE", lineno=3),
                ]
            )
            code = ControlFlowGraph.from_bytecode(code)

            end_loop = Label()
            self.check(
                code,
                Instr("SETUP_LOOP", end_loop, lineno=2),
                Instr("LOAD_CONST", 7, lineno=3),
                Instr("RETURN_VALUE", lineno=3),
                end_loop,
                Instr("LOAD_CONST", None, lineno=3),
                Instr("RETURN_VALUE", lineno=3),
            )
        else:
            setup_loop = Label()
            return_label = Label()
            code = Bytecode(
                [
                    setup_loop,
                    Instr("LOAD_CONST", 7, lineno=3),
                    Instr("RETURN_VALUE", lineno=3),
                    Instr("JUMP_ABSOLUTE", setup_loop, lineno=3),
                    Instr("LOAD_CONST", None, lineno=3),
                    Instr("RETURN_VALUE", lineno=3),
                ]
            )
            code = ControlFlowGraph.from_bytecode(code)

            self.check(code, Instr("LOAD_CONST", 7, lineno=3), Instr("RETURN_VALUE", lineno=3))

    def test_not_jump_if_false(self):
        # Replace UNARY_NOT+POP_JUMP_IF_FALSE with POP_JUMP_IF_TRUE
        #
        # if not x:
        #     y = 9
        label = Label()
        code = Bytecode(
            [
                Instr("LOAD_NAME", "x"),
                Instr("UNARY_NOT"),
                Instr("POP_JUMP_IF_FALSE", label),
                Instr("LOAD_CONST", 9),
                Instr("STORE_NAME", "y"),
                label,
            ]
        )

        code = self.optimize_blocks(code)
        label = Label()
        self.check(
            code,
            Instr("LOAD_NAME", "x"),
            Instr("POP_JUMP_IF_TRUE", label),
            Instr("LOAD_CONST", 9),
            Instr("STORE_NAME", "y"),
            label,
        )

    def test_unconditional_jump_to_return(self):
        # def func():
        #     if test:
        #         if test2:
        #             x = 10
        #         else:
        #             x = 20
        #     else:
        #         x = 30

        label_instr11 = Label()
        label_instr14 = Label()
        label_instr7 = Label()
        code = Bytecode(
            [
                Instr("LOAD_GLOBAL", "test", lineno=2),
                Instr("POP_JUMP_IF_FALSE", label_instr11, lineno=2),
                Instr("LOAD_GLOBAL", "test2", lineno=3),
                Instr("POP_JUMP_IF_FALSE", label_instr7, lineno=3),
                Instr("LOAD_CONST", 10, lineno=4),
                Instr("STORE_FAST", "x", lineno=4),
                Instr("JUMP_ABSOLUTE", label_instr14, lineno=4),
                label_instr7,
                Instr("LOAD_CONST", 20, lineno=6),
                Instr("STORE_FAST", "x", lineno=6),
                Instr("JUMP_FORWARD", label_instr14, lineno=6),
                label_instr11,
                Instr("LOAD_CONST", 30, lineno=8),
                Instr("STORE_FAST", "x", lineno=8),
                label_instr14,
                Instr("LOAD_CONST", None, lineno=8),
                Instr("RETURN_VALUE", lineno=8),
            ]
        )

        label1 = Label()
        label3 = Label()
        label4 = Label()
        self.check(
            code,
            Instr("LOAD_GLOBAL", "test", lineno=2),
            Instr("POP_JUMP_IF_FALSE", label3, lineno=2),
            Instr("LOAD_GLOBAL", "test2", lineno=3),
            Instr("POP_JUMP_IF_FALSE", label1, lineno=3),
            Instr("LOAD_CONST", 10, lineno=4),
            Instr("STORE_FAST", "x", lineno=4),
            Instr("JUMP_ABSOLUTE", label4, lineno=4),
            label1,
            Instr("LOAD_CONST", 20, lineno=6),
            Instr("STORE_FAST", "x", lineno=6),
            Instr("JUMP_FORWARD", label4, lineno=6),
            label3,
            Instr("LOAD_CONST", 30, lineno=8),
            Instr("STORE_FAST", "x", lineno=8),
            label4,
            Instr("LOAD_CONST", None, lineno=8),
            Instr("RETURN_VALUE", lineno=8),
        )

    def test_unconditional_jumps(self):
        # def func():
        #     if x:
        #         if y:
        #             func()
        label_instr7 = Label()
        code = Bytecode(
            [
                Instr("LOAD_GLOBAL", "x", lineno=2),
                Instr("POP_JUMP_IF_FALSE", label_instr7, lineno=2),
                Instr("LOAD_GLOBAL", "y", lineno=3),
                Instr("POP_JUMP_IF_FALSE", label_instr7, lineno=3),
                Instr("LOAD_GLOBAL", "func", lineno=4),
                Instr("CALL_FUNCTION", 0, lineno=4),
                Instr("POP_TOP", lineno=4),
                label_instr7,
                Instr("LOAD_CONST", None, lineno=4),
                Instr("RETURN_VALUE", lineno=4),
            ]
        )

        label_return = Label()
        self.check(
            code,
            Instr("LOAD_GLOBAL", "x", lineno=2),
            Instr("POP_JUMP_IF_FALSE", label_return, lineno=2),
            Instr("LOAD_GLOBAL", "y", lineno=3),
            Instr("POP_JUMP_IF_FALSE", label_return, lineno=3),
            Instr("LOAD_GLOBAL", "func", lineno=4),
            Instr("CALL_FUNCTION", 0, lineno=4),
            Instr("POP_TOP", lineno=4),
            label_return,
            Instr("LOAD_CONST", None, lineno=4),
            Instr("RETURN_VALUE", lineno=4),
        )

    def test_jump_to_return(self):
        # def func(condition):
        #     return 'yes' if condition else 'no'
        label_instr4 = Label()
        label_instr6 = Label()
        code = Bytecode(
            [
                Instr("LOAD_FAST", "condition"),
                Instr("POP_JUMP_IF_FALSE", label_instr4),
                Instr("LOAD_CONST", "yes"),
                Instr("JUMP_FORWARD", label_instr6),
                label_instr4,
                Instr("LOAD_CONST", "no"),
                label_instr6,
                Instr("RETURN_VALUE"),
            ]
        )

        label = Label()
        self.check(
            code,
            Instr("LOAD_FAST", "condition"),
            Instr("POP_JUMP_IF_FALSE", label),
            Instr("LOAD_CONST", "yes"),
            Instr("RETURN_VALUE"),
            label,
            Instr("LOAD_CONST", "no"),
            Instr("RETURN_VALUE"),
        )

    def test_jump_if_true_to_jump_if_false(self):
        # Replace JUMP_IF_TRUE_OR_POP jumping to POP_JUMP_IF_FALSE <target>
        # with POP_JUMP_IF_TRUE <offset after the second POP_JUMP_IF_FALSE>
        #
        #     if x or y:
        #         z = 1

        label_instr3 = Label()
        label_instr7 = Label()
        code = Bytecode(
            [
                Instr("LOAD_NAME", "x"),
                Instr("JUMP_IF_TRUE_OR_POP", label_instr3),
                Instr("LOAD_NAME", "y"),
                label_instr3,
                Instr("POP_JUMP_IF_FALSE", label_instr7),
                Instr("LOAD_CONST", 1),
                Instr("STORE_NAME", "z"),
                label_instr7,
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )

        label_instr4 = Label()
        label_instr7 = Label()
        self.check(
            code,
            Instr("LOAD_NAME", "x"),
            Instr("POP_JUMP_IF_TRUE", label_instr4),
            Instr("LOAD_NAME", "y"),
            Instr("POP_JUMP_IF_FALSE", label_instr7),
            label_instr4,
            Instr("LOAD_CONST", 1),
            Instr("STORE_NAME", "z"),
            label_instr7,
            Instr("LOAD_CONST", None),
            Instr("RETURN_VALUE"),
        )

    def test_jump_if_false_to_jump_if_false(self):
        # Replace JUMP_IF_FALSE_OR_POP jumping to POP_JUMP_IF_FALSE <label>
        # with POP_JUMP_IF_FALSE <label>
        #
        #     while n > 0 and start > 3:
        #         func()
        if sys.version_info < (3, 8):
            label_instr1 = Label()
            label_instr15 = Label()
            label_instr17 = Label()
            label_instr9 = Label()
            code = Bytecode(
                [
                    Instr("SETUP_LOOP", label_instr17),
                    label_instr1,
                    Instr("LOAD_NAME", "n"),
                    Instr("LOAD_CONST", 0),
                    Instr("COMPARE_OP", Compare.GT),
                    # JUMP_IF_FALSE_OR_POP jumps to POP_JUMP_IF_FALSE
                    # which jumps to label_instr15
                    Instr("JUMP_IF_FALSE_OR_POP", label_instr9),
                    Instr("LOAD_NAME", "start"),
                    Instr("LOAD_CONST", 3),
                    Instr("COMPARE_OP", Compare.GT),
                    label_instr9,
                    Instr("POP_JUMP_IF_FALSE", label_instr15),
                    Instr("LOAD_NAME", "func"),
                    Instr("CALL_FUNCTION", 0),
                    Instr("POP_TOP"),
                    Instr("JUMP_ABSOLUTE", label_instr1),
                    label_instr15,
                    Instr("POP_BLOCK"),
                    label_instr17,
                    Instr("LOAD_CONST", None),
                    Instr("RETURN_VALUE"),
                ]
            )

            label_instr1 = Label()
            label_instr14 = Label()
            label_instr16 = Label()
            self.check(
                code,
                Instr("SETUP_LOOP", label_instr16),
                label_instr1,
                Instr("LOAD_NAME", "n"),
                Instr("LOAD_CONST", 0),
                Instr("COMPARE_OP", Compare.GT),
                Instr("POP_JUMP_IF_FALSE", label_instr14),
                Instr("LOAD_NAME", "start"),
                Instr("LOAD_CONST", 3),
                Instr("COMPARE_OP", Compare.GT),
                Instr("POP_JUMP_IF_FALSE", label_instr14),
                Instr("LOAD_NAME", "func"),
                Instr("CALL_FUNCTION", 0),
                Instr("POP_TOP"),
                Instr("JUMP_ABSOLUTE", label_instr1),
                label_instr14,
                Instr("POP_BLOCK"),
                label_instr16,
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            )
        else:
            label_instr1 = Label()
            label_instr15 = Label()
            label_instr9 = Label()
            code = Bytecode(
                [
                    label_instr1,
                    Instr("LOAD_NAME", "n"),
                    Instr("LOAD_CONST", 0),
                    Instr("COMPARE_OP", Compare.GT),
                    # JUMP_IF_FALSE_OR_POP jumps to POP_JUMP_IF_FALSE
                    # which jumps to label_instr15
                    Instr("JUMP_IF_FALSE_OR_POP", label_instr9),
                    Instr("LOAD_NAME", "start"),
                    Instr("LOAD_CONST", 3),
                    Instr("COMPARE_OP", Compare.GT),
                    label_instr9,
                    Instr("POP_JUMP_IF_FALSE", label_instr15),
                    Instr("LOAD_NAME", "func"),
                    Instr("CALL_FUNCTION", 0),
                    Instr("POP_TOP"),
                    Instr("JUMP_ABSOLUTE", label_instr1),
                    label_instr15,
                    Instr("LOAD_CONST", None),
                    Instr("RETURN_VALUE"),
                ]
            )

            label_instr1 = Label()
            label_instr14 = Label()
            self.check(
                code,
                label_instr1,
                Instr("LOAD_NAME", "n"),
                Instr("LOAD_CONST", 0),
                Instr("COMPARE_OP", Compare.GT),
                Instr("POP_JUMP_IF_FALSE", label_instr14),
                Instr("LOAD_NAME", "start"),
                Instr("LOAD_CONST", 3),
                Instr("COMPARE_OP", Compare.GT),
                Instr("POP_JUMP_IF_FALSE", label_instr14),
                Instr("LOAD_NAME", "func"),
                Instr("CALL_FUNCTION", 0),
                Instr("POP_TOP"),
                Instr("JUMP_ABSOLUTE", label_instr1),
                label_instr14,
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            )

    def test_nop(self):
        code = Bytecode([Instr("LOAD_NAME", "x"), Instr("NOP"), Instr("STORE_NAME", "test")])

        self.check(code, Instr("LOAD_NAME", "x"), Instr("STORE_NAME", "test"))

    def test_dead_code_jump(self):
        label = Label()
        code = Bytecode(
            [
                Instr("LOAD_NAME", "x"),
                Instr("JUMP_ABSOLUTE", label),
                # dead code
                Instr("LOAD_NAME", "y"),
                Instr("STORE_NAME", "test"),
                label,
                Instr("STORE_NAME", "test"),
            ]
        )

        self.check(code, Instr("LOAD_NAME", "x"), Instr("STORE_NAME", "test"))

    def test_uncond_jump_to_uncond_jump(self):
        # Replace JUMP_FORWARD t1 jumping to JUMP_FORWARD t2
        # with JUMP_ABSOLUTE t2

        label = Label()
        label2 = Label()
        label3 = Label()
        label4 = Label()
        code = Bytecode(
            [
                Instr("LOAD_NAME", "test"),
                Instr("POP_JUMP_IF_TRUE", label),
                # redundant jump
                Instr("JUMP_FORWARD", label2),
                label,
                Instr("LOAD_CONST", 1),
                Instr("STORE_NAME", "x"),
                Instr("LOAD_NAME", "test"),
                Instr("POP_JUMP_IF_TRUE", label3),
                label2,
                Instr("JUMP_FORWARD", label4),
                label3,
                Instr("LOAD_CONST", 1),
                Instr("STORE_NAME", "x"),
                label4,
                Instr("LOAD_CONST", None),
                Instr("RETURN_VALUE"),
            ]
        )

        label = Label()
        label3 = Label()
        label4 = Label()
        self.check(
            code,
            Instr("LOAD_NAME", "test"),
            Instr("POP_JUMP_IF_TRUE", label),
            # JUMP_FORWARD label2 was replaced with JUMP_ABSOLUTE label4
            Instr("JUMP_ABSOLUTE", label4),
            label,
            Instr("LOAD_CONST", 1),
            Instr("STORE_NAME", "x"),
            Instr("LOAD_NAME", "test"),
            Instr("POP_JUMP_IF_TRUE", label3),
            Instr("JUMP_FORWARD", label4),
            label3,
            Instr("LOAD_CONST", 1),
            Instr("STORE_NAME", "x"),
            label4,
            Instr("LOAD_CONST", None),
            Instr("RETURN_VALUE"),
        )


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



import subprocess
import sys

print(sys.executable)

if __name__ == "__main__":
    p = subprocess.Popen([sys.executable, "-u", "_always_live_program.py"])
    import attach_pydevd

    attach_pydevd.main(attach_pydevd.process_command_line(["--pid", str(p.pid), "--protocol", "http"]))
    p.wait()



import sys
import struct

print("Executable: %s" % sys.executable)
import os


def loop_in_thread():
    while True:
        import time

        time.sleep(0.5)
        sys.stdout.write("#")
        sys.stdout.flush()


import threading

threading.Thread(target=loop_in_thread).start()


def is_python_64bit():
    return struct.calcsize("P") == 8


print("Is 64: %s" % is_python_64bit())

if __name__ == "__main__":
    print("pid:%s" % (os.getpid()))
    i = 0
    while True:
        i += 1
        import time

        time.sleep(0.5)
        sys.stdout.write(".")
        sys.stdout.flush()
        if i % 40 == 0:
            sys.stdout.write("\n")
            sys.stdout.flush()



import json
import urllib.parse as urllib_parse


class DebugOptions(object):
    __slots__ = [
        "just_my_code",
        "redirect_output",
        "show_return_value",
        "break_system_exit_zero",
        "django_debug",
        "flask_debug",
        "stop_on_entry",
        "max_exception_stack_frames",
        "gui_event_loop",
        "client_os",
    ]

    def __init__(self):
        self.just_my_code = True
        self.redirect_output = False
        self.show_return_value = False
        self.break_system_exit_zero = False
        self.django_debug = False
        self.flask_debug = False
        self.stop_on_entry = False
        self.max_exception_stack_frames = 0
        self.gui_event_loop = "matplotlib"
        self.client_os = None

    def to_json(self):
        dct = {}
        for s in self.__slots__:
            dct[s] = getattr(self, s)
        return json.dumps(dct)

    def update_fom_debug_options(self, debug_options):
        if "DEBUG_STDLIB" in debug_options:
            self.just_my_code = not debug_options.get("DEBUG_STDLIB")

        if "REDIRECT_OUTPUT" in debug_options:
            self.redirect_output = debug_options.get("REDIRECT_OUTPUT")

        if "SHOW_RETURN_VALUE" in debug_options:
            self.show_return_value = debug_options.get("SHOW_RETURN_VALUE")

        if "BREAK_SYSTEMEXIT_ZERO" in debug_options:
            self.break_system_exit_zero = debug_options.get("BREAK_SYSTEMEXIT_ZERO")

        if "DJANGO_DEBUG" in debug_options:
            self.django_debug = debug_options.get("DJANGO_DEBUG")

        if "FLASK_DEBUG" in debug_options:
            self.flask_debug = debug_options.get("FLASK_DEBUG")

        if "STOP_ON_ENTRY" in debug_options:
            self.stop_on_entry = debug_options.get("STOP_ON_ENTRY")

        if "CLIENT_OS_TYPE" in debug_options:
            self.client_os = debug_options.get("CLIENT_OS_TYPE")

        # Note: _max_exception_stack_frames cannot be set by debug options.

    def update_from_args(self, args):
        if "justMyCode" in args:
            self.just_my_code = bool_parser(args["justMyCode"])
        else:
            # i.e.: if justMyCode is provided, don't check the deprecated value
            if "debugStdLib" in args:
                self.just_my_code = not bool_parser(args["debugStdLib"])

        if "redirectOutput" in args:
            self.redirect_output = bool_parser(args["redirectOutput"])

        if "showReturnValue" in args:
            self.show_return_value = bool_parser(args["showReturnValue"])

        if "breakOnSystemExitZero" in args:
            self.break_system_exit_zero = bool_parser(args["breakOnSystemExitZero"])

        if "django" in args:
            self.django_debug = bool_parser(args["django"])

        if "flask" in args:
            self.flask_debug = bool_parser(args["flask"])

        if "jinja" in args:
            self.flask_debug = bool_parser(args["jinja"])

        if "stopOnEntry" in args:
            self.stop_on_entry = bool_parser(args["stopOnEntry"])

        self.max_exception_stack_frames = int_parser(args.get("maxExceptionStackFrames", 0))

        if "guiEventLoop" in args:
            self.gui_event_loop = str(args["guiEventLoop"])

        if "clientOS" in args:
            self.client_os = str(args["clientOS"]).upper()


def int_parser(s, default_value=0):
    try:
        return int(s)
    except Exception:
        return default_value


def bool_parser(s):
    return s in ("True", "true", "1", True, 1)


def unquote(s):
    return None if s is None else urllib_parse.unquote(s)


DEBUG_OPTIONS_PARSER = {
    "WAIT_ON_ABNORMAL_EXIT": bool_parser,
    "WAIT_ON_NORMAL_EXIT": bool_parser,
    "BREAK_SYSTEMEXIT_ZERO": bool_parser,
    "REDIRECT_OUTPUT": bool_parser,
    "DJANGO_DEBUG": bool_parser,
    "FLASK_DEBUG": bool_parser,
    "FIX_FILE_PATH_CASE": bool_parser,
    "CLIENT_OS_TYPE": unquote,
    "DEBUG_STDLIB": bool_parser,
    "STOP_ON_ENTRY": bool_parser,
    "SHOW_RETURN_VALUE": bool_parser,
    "MULTIPROCESS": bool_parser,
}

DEBUG_OPTIONS_BY_FLAG = {
    "RedirectOutput": "REDIRECT_OUTPUT=True",
    "WaitOnNormalExit": "WAIT_ON_NORMAL_EXIT=True",
    "WaitOnAbnormalExit": "WAIT_ON_ABNORMAL_EXIT=True",
    "BreakOnSystemExitZero": "BREAK_SYSTEMEXIT_ZERO=True",
    "Django": "DJANGO_DEBUG=True",
    "Flask": "FLASK_DEBUG=True",
    "Jinja": "FLASK_DEBUG=True",
    "FixFilePathCase": "FIX_FILE_PATH_CASE=True",
    "DebugStdLib": "DEBUG_STDLIB=True",
    "WindowsClient": "CLIENT_OS_TYPE=WINDOWS",
    "UnixClient": "CLIENT_OS_TYPE=UNIX",
    "StopOnEntry": "STOP_ON_ENTRY=True",
    "ShowReturnValue": "SHOW_RETURN_VALUE=True",
    "Multiprocess": "MULTIPROCESS=True",
}


def _build_debug_options(flags):
    """Build string representation of debug options from the launch config."""
    return ";".join(DEBUG_OPTIONS_BY_FLAG[flag] for flag in flags or [] if flag in DEBUG_OPTIONS_BY_FLAG)


def _parse_debug_options(opts):
    """Debug options are semicolon separated key=value pairs"""
    options = {}
    if not opts:
        return options

    for opt in opts.split(";"):
        try:
            key, value = opt.split("=")
        except ValueError:
            continue
        try:
            options[key] = DEBUG_OPTIONS_PARSER[key](value)
        except KeyError:
            continue

    return options


def _extract_debug_options(opts, flags=None):
    """Return the debug options encoded in the given value.

    "opts" is a semicolon-separated string of "key=value" pairs.
    "flags" is a list of strings.

    If flags is provided then it is used as a fallback.

    The values come from the launch config:

     {
         type:'python',
         request:'launch'|'attach',
         name:'friendly name for debug config',
         debugOptions:[
             'RedirectOutput', 'Django'
         ],
         options:'REDIRECT_OUTPUT=True;DJANGO_DEBUG=True'
     }

    Further information can be found here:

    https://code.visualstudio.com/docs/editor/debugging#_launchjson-attributes
    """
    if not opts:
        opts = _build_debug_options(flags)
    return _parse_debug_options(opts)



#!/usr/bin/env python

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

import os
import os.path
import setuptools
import subprocess
import sys


sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import versioneer  # noqa

del sys.path[0]

sys.path.insert(0, os.path.join(os.path.dirname(os.path.abspath(__file__)), "src"))
import debugpy
import debugpy._vendored

del sys.path[0]


PYDEVD_ROOT = debugpy._vendored.project_root("pydevd")
DEBUGBY_ROOT = os.path.dirname(os.path.abspath(debugpy.__file__))


def get_buildplatform():
    if "-p" in sys.argv:
        return sys.argv[sys.argv.index("-p") + 1]
    return None


# bdist_wheel determines whether the package is pure or not based on ext_modules.
# However, all pydevd native modules are prebuilt and packaged as data, so they
# should not be in the list.
#
# The proper way to handle this is by overriding has_ext_modules. However, due to
# https://bugs.python.org/issue32957, in setuptools 57.0.0 and below, it is not
# always called when it should be, with ext_modules tested directly instead.
#
# So, for non-pure builds, we provide a customized empty list for ext_modules that
# tests as truthful - this causes the package to be treated as non-pure on all
# relevant setuptools versions.
class ExtModules(list):
    def __bool__(self):
        return True


def override_build(cmds):
    def finalize_options(self):
        # Mark all packages as pure if requested to build a universal wheel.
        # Otherwise, bdist_wheel will silently ignore the request.
        bdist_wheel = self.distribution.get_command_obj("bdist_wheel", 0)
        if bdist_wheel is not None and bdist_wheel.universal:
            self.distribution.ext_modules = []
            self.distribution.has_ext_modules = lambda: False
        original(self)

    try:
        build = cmds["build"]
    except KeyError:
        from distutils.command.build import build
    original = build.finalize_options
    build.finalize_options = finalize_options


def override_build_py(cmds):
    # Add vendored packages as data files for this package, and filter platform-specific binaries
    # in data files for binary builds.
    def finalize_options(self):
        original(self)

        # Ensure that pydevd extensions are present for inclusion into data_files.
        self.announce(
            "Compiling pydevd Cython extension modules (set SKIP_CYTHON_BUILD=1 to omit)."
        )
        try:
            subprocess.check_call([
                sys.executable,
                os.path.join(PYDEVD_ROOT, "setup_pydevd_cython.py"),
                "build_ext",
                "--inplace",
            ])
        except subprocess.SubprocessError:
            # pydevd Cython extensions are optional performance enhancements, and debugpy is
            # usable without them. Thus, we want to ignore build errors here by default, so
            # that a user can do `pip install debugpy` on a platform for which we don't supply
            # prebuild wheels even if they don't have the C build tools installed. On the other
            # hand, for our own official builds and for local development, we want to know if
            # the extensions fail to build.
            if int(os.getenv("REQUIRE_CYTHON_BUILD", "0")):
                raise
            self.warn(
                "Failed to compile pydevd Cython extension modules (set SKIP_CYTHON_BUILD=1 to omit); proceeding without them."
            )

        # Register pydevd and other vendored packages as package data for debugpy.
        vendored_files = self.package_data["debugpy._vendored"]
        for project in debugpy._vendored.list_all():
            for filename in debugpy._vendored.iter_packaging_files(project):
                vendored_files.append(filename)

        # Don't filter anything for universal wheels.
        if not self.distribution.ext_modules:
            return

        plat = self.get_finalized_command("build").plat_name

        def is_applicable(filename):
            def tail_is(*suffixes):
                return any((filename.endswith(ext) for ext in suffixes))

            if tail_is(".dylib"):
                return plat.startswith("macosx")
            if tail_is(".exe", ".dll", ".pdb", ".pyd"):
                return plat in ("win32", "win-amd64")
            if tail_is("-i386-linux-gnu.so", "_linux_x86.so"):
                return plat == "linux-i686"
            if tail_is("-x86_64-linux-gnu.so", "_linux_amd64.so"):
                return plat == "linux-x86_64"
            return True

        vendored_files[:] = list(filter(is_applicable, vendored_files))

    try:
        build_py = cmds["build_py"]
    except KeyError:
        from setuptools.command.build_py import build_py
    original = build_py.finalize_options
    build_py.finalize_options = finalize_options


with open("DESCRIPTION.md", "r") as fh:
    long_description = fh.read()


if __name__ == "__main__":
    extras = {}
    platforms = get_buildplatform()
    if platforms is not None:
        extras["platforms"] = platforms

    cmds = versioneer.get_cmdclass()
    override_build(cmds)
    override_build_py(cmds)

    setuptools.setup(
        name="debugpy",
        version=versioneer.get_version(),
        description="An implementation of the Debug Adapter Protocol for Python",  # noqa
        long_description=long_description,
        long_description_content_type="text/markdown",
        license="MIT",
        author="Microsoft Corporation",
        author_email="ptvshelp@microsoft.com",
        url="https://aka.ms/debugpy",
        project_urls={
            "Source": "https://github.com/microsoft/debugpy",
        },
        python_requires=">=3.8",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Python :: 3.13",
            "Programming Language :: Python :: 3.14",
            "Topic :: Software Development :: Debuggers",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: MacOS",
            "Operating System :: POSIX",
            "License :: OSI Approved :: MIT License",
        ],
        package_dir={"": "src"},
        packages=[
            "debugpy",
            "debugpy.adapter",
            "debugpy.common",
            "debugpy.launcher",
            "debugpy.server",
            "debugpy._vendored",
        ],
        package_data={
            "debugpy": ["ThirdPartyNotices.txt", "py.typed"],
            "debugpy._vendored": [
                # pydevd extensions must be built before this list can be computed properly,
                # so it is populated in the overridden build_py.finalize_options().
            ],
        },
        ext_modules=ExtModules(),
        has_ext_modules=lambda: True,
        cmdclass=cmds,
        # allow the user to call "debugpy" instead of "python -m debugpy"
        entry_points={
            "console_scripts": [
                "debugpy = debugpy.server.cli:main",
                "debugpy-adapter = debugpy.adapter.__main__:main",
            ],
        },
        **extras,
    )



#!~/.wine/drive_c/Python25/python.exe
# -*- coding: utf-8 -*-

# Copyright (c) 2009-2014, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Debugging.

@group Debugging:
    Debug

@group Warnings:
    MixedBitsWarning
"""

__revision__ = "$Id$"

__all__ = ["Debug", "MixedBitsWarning"]

import sys
from winappdbg import win32
from winappdbg.system import System
from winappdbg.process import Process
from winappdbg.thread import Thread
from winappdbg.module import Module
from winappdbg.window import Window
from winappdbg.breakpoint import _BreakpointContainer, CodeBreakpoint
from winappdbg.event import Event, EventHandler, EventDispatcher, EventFactory
from winappdbg.interactive import ConsoleDebugger

import warnings
##import traceback

# ==============================================================================


# If you set this warning to be considered as an error, you can stop the
# debugger from attaching to 64-bit processes from a 32-bit Python VM and
# visceversa.
class MixedBitsWarning(RuntimeWarning):
    """
    This warning is issued when mixing 32 and 64 bit processes.
    """


# ==============================================================================

# TODO
# * Add memory read and write operations, similar to those in the Process
#   class, but hiding the presence of the code breakpoints.
# * Add a method to get the memory map of a process, but hiding the presence
#   of the page breakpoints.
# * Maybe the previous two features should be implemented at the Process class
#   instead, but how to communicate with the Debug object without creating
#   circular references? Perhaps the "overrides" could be set using private
#   members (so users won't see them), but then there's the problem of the
#   users being able to access the snapshot (i.e. clear it), which is why it's
#   not such a great idea to use the snapshot to store data that really belongs
#   to the Debug class.


class Debug(EventDispatcher, _BreakpointContainer):
    """
    The main debugger class.

    @group Debugging:
        interactive, attach, detach, detach_from_all, execv, execl,
        kill, kill_all,
        get_debugee_count, get_debugee_pids,
        is_debugee, is_debugee_attached, is_debugee_started,
        in_hostile_mode,
        add_existing_session

    @group Debugging loop:
        loop, stop, next, wait, dispatch, cont

    @undocumented: force_garbage_collection

    @type system: L{System}
    @ivar system: A System snapshot that is automatically updated for
        processes being debugged. Processes not being debugged in this snapshot
        may be outdated.
    """

    # Automatically set to True the first time a Debug object is instanced.
    _debug_static_init = False

    def __init__(self, eventHandler=None, bKillOnExit=False, bHostileCode=False):
        """
        Debugger object.

        @type  eventHandler: L{EventHandler}
        @param eventHandler:
            (Optional, recommended) Custom event handler object.

        @type  bKillOnExit: bool
        @param bKillOnExit: (Optional) Kill on exit mode.
            If C{True} debugged processes are killed when the debugger is
            stopped. If C{False} when the debugger stops it detaches from all
            debugged processes and leaves them running (default).

        @type  bHostileCode: bool
        @param bHostileCode: (Optional) Hostile code mode.
            Set to C{True} to take some basic precautions against anti-debug
            tricks. Disabled by default.

        @warn: When hostile mode is enabled, some things may not work as
            expected! This is because the anti-anti debug tricks may disrupt
            the behavior of the Win32 debugging APIs or WinAppDbg itself.

        @note: The L{eventHandler} parameter may be any callable Python object
            (for example a function, or an instance method).
            However you'll probably find it more convenient to use an instance
            of a subclass of L{EventHandler} here.

        @raise WindowsError: Raises an exception on error.
        """
        EventDispatcher.__init__(self, eventHandler)
        _BreakpointContainer.__init__(self)

        self.system = System()
        self.lastEvent = None
        self.__firstDebugee = True
        self.__bKillOnExit = bKillOnExit
        self.__bHostileCode = bHostileCode
        self.__breakOnEP = set()  # set of pids
        self.__attachedDebugees = set()  # set of pids
        self.__startedDebugees = set()  # set of pids

        if not self._debug_static_init:
            self._debug_static_init = True

            # Request debug privileges for the current process.
            # Only do this once, and only after instancing a Debug object,
            # so passive debuggers don't get detected because of this.
            self.system.request_debug_privileges(bIgnoreExceptions=False)

            # Try to fix the symbol store path if it wasn't set.
            # But don't enable symbol downloading by default, since it may
            # degrade performance severely.
            self.system.fix_symbol_store_path(remote=False, force=False)

    ##    # It's hard not to create circular references,
    ##    # and if we have a destructor, we can end up leaking everything.
    ##    # It's best to code the debugging loop properly to always
    ##    # stop the debugger before going out of scope.
    ##    def __del__(self):
    ##        self.stop()

    def __enter__(self):
        """
        Compatibility with the "C{with}" Python statement.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        Compatibility with the "C{with}" Python statement.
        """
        self.stop()

    def __len__(self):
        """
        @rtype:  int
        @return: Number of processes being debugged.
        """
        return self.get_debugee_count()

    # TODO: maybe custom __bool__ to break out of loop() ?
    # it already does work (because of __len__) but it'd be
    # useful to do it from the event handler anyway

    # ------------------------------------------------------------------------------

    def __setSystemKillOnExitMode(self):
        # Make sure the default system behavior on detaching from processes
        # versus killing them matches our preferences. This only affects the
        # scenario where the Python VM dies unexpectedly without running all
        # the finally clauses, or the user failed to either instance the Debug
        # object inside a with block or call the stop() method before quitting.
        if self.__firstDebugee:
            try:
                System.set_kill_on_exit_mode(self.__bKillOnExit)
                self.__firstDebugee = False
            except Exception:
                pass

    def attach(self, dwProcessId):
        """
        Attaches to an existing process for debugging.

        @see: L{detach}, L{execv}, L{execl}

        @type  dwProcessId: int
        @param dwProcessId: Global ID of a process to attach to.

        @rtype:  L{Process}
        @return: A new Process object. Normally you don't need to use it now,
            it's best to interact with the process from the event handler.

        @raise WindowsError: Raises an exception on error.
            Depending on the circumstances, the debugger may or may not have
            attached to the target process.
        """

        # Get the Process object from the snapshot,
        # if missing create a new one.
        try:
            aProcess = self.system.get_process(dwProcessId)
        except KeyError:
            aProcess = Process(dwProcessId)

        # Warn when mixing 32 and 64 bits.
        # This also allows the user to stop attaching altogether,
        # depending on how the warnings are configured.
        if System.bits != aProcess.get_bits():
            msg = "Mixture of 32 and 64 bits is considered experimental." " Use at your own risk!"
            warnings.warn(msg, MixedBitsWarning)

        # Attach to the process.
        win32.DebugActiveProcess(dwProcessId)

        # Add the new PID to the set of debugees.
        self.__attachedDebugees.add(dwProcessId)

        # Match the system kill-on-exit flag to our own.
        self.__setSystemKillOnExitMode()

        # If the Process object was not in the snapshot, add it now.
        if not self.system.has_process(dwProcessId):
            self.system._add_process(aProcess)

        # Scan the process threads and loaded modules.
        # This is prefered because the thread and library events do not
        # properly give some information, like the filename for each module.
        aProcess.scan_threads()
        aProcess.scan_modules()

        # Return the Process object, like the execv() and execl() methods.
        return aProcess

    def execv(self, argv, **kwargs):
        """
        Starts a new process for debugging.

        This method uses a list of arguments. To use a command line string
        instead, use L{execl}.

        @see: L{attach}, L{detach}

        @type  argv: list( str... )
        @param argv: List of command line arguments to pass to the debugee.
            The first element must be the debugee executable filename.

        @type    bBreakOnEntryPoint: bool
        @keyword bBreakOnEntryPoint: C{True} to automatically set a breakpoint
            at the program entry point.

        @type    bConsole: bool
        @keyword bConsole: True to inherit the console of the debugger.
            Defaults to C{False}.

        @type    bFollow: bool
        @keyword bFollow: C{True} to automatically attach to child processes.
            Defaults to C{False}.

        @type    bInheritHandles: bool
        @keyword bInheritHandles: C{True} if the new process should inherit
            it's parent process' handles. Defaults to C{False}.

        @type    bSuspended: bool
        @keyword bSuspended: C{True} to suspend the main thread before any code
            is executed in the debugee. Defaults to C{False}.

        @keyword dwParentProcessId: C{None} or C{0} if the debugger process
            should be the parent process (default), or a process ID to
            forcefully set as the debugee's parent (only available for Windows
            Vista and above).

            In hostile mode, the default is not the debugger process but the
            process ID for "explorer.exe".

        @type    iTrustLevel: int or None
        @keyword iTrustLevel: Trust level.
            Must be one of the following values:
             - 0: B{No trust}. May not access certain resources, such as
                  cryptographic keys and credentials. Only available since
                  Windows XP and 2003, desktop editions. This is the default
                  in hostile mode.
             - 1: B{Normal trust}. Run with the same privileges as a normal
                  user, that is, one that doesn't have the I{Administrator} or
                  I{Power User} user rights. Only available since Windows XP
                  and 2003, desktop editions.
             - 2: B{Full trust}. Run with the exact same privileges as the
                  current user. This is the default in normal mode.

        @type    bAllowElevation: bool
        @keyword bAllowElevation: C{True} to allow the child process to keep
            UAC elevation, if the debugger itself is running elevated. C{False}
            to ensure the child process doesn't run with elevation. Defaults to
            C{True}.

            This flag is only meaningful on Windows Vista and above, and if the
            debugger itself is running with elevation. It can be used to make
            sure the child processes don't run elevated as well.

            This flag DOES NOT force an elevation prompt when the debugger is
            not running with elevation.

            Note that running the debugger with elevation (or the Python
            interpreter at all for that matter) is not normally required.
            You should only need to if the target program requires elevation
            to work properly (for example if you try to debug an installer).

        @rtype:  L{Process}
        @return: A new Process object. Normally you don't need to use it now,
            it's best to interact with the process from the event handler.

        @raise WindowsError: Raises an exception on error.
        """
        if type(argv) in (str, compat.unicode):
            raise TypeError("Debug.execv expects a list, not a string")
        lpCmdLine = self.system.argv_to_cmdline(argv)
        return self.execl(lpCmdLine, **kwargs)

    def execl(self, lpCmdLine, **kwargs):
        """
        Starts a new process for debugging.

        This method uses a command line string. To use a list of arguments
        instead, use L{execv}.

        @see: L{attach}, L{detach}

        @type  lpCmdLine: str
        @param lpCmdLine: Command line string to execute.
            The first token must be the debugee executable filename.
            Tokens with spaces must be enclosed in double quotes.
            Tokens including double quote characters must be escaped with a
            backslash.

        @type    bBreakOnEntryPoint: bool
        @keyword bBreakOnEntryPoint: C{True} to automatically set a breakpoint
            at the program entry point. Defaults to C{False}.

        @type    bConsole: bool
        @keyword bConsole: True to inherit the console of the debugger.
            Defaults to C{False}.

        @type    bFollow: bool
        @keyword bFollow: C{True} to automatically attach to child processes.
            Defaults to C{False}.

        @type    bInheritHandles: bool
        @keyword bInheritHandles: C{True} if the new process should inherit
            it's parent process' handles. Defaults to C{False}.

        @type    bSuspended: bool
        @keyword bSuspended: C{True} to suspend the main thread before any code
            is executed in the debugee. Defaults to C{False}.

        @type    dwParentProcessId: int or None
        @keyword dwParentProcessId: C{None} or C{0} if the debugger process
            should be the parent process (default), or a process ID to
            forcefully set as the debugee's parent (only available for Windows
            Vista and above).

            In hostile mode, the default is not the debugger process but the
            process ID for "explorer.exe".

        @type    iTrustLevel: int
        @keyword iTrustLevel: Trust level.
            Must be one of the following values:
             - 0: B{No trust}. May not access certain resources, such as
                  cryptographic keys and credentials. Only available since
                  Windows XP and 2003, desktop editions. This is the default
                  in hostile mode.
             - 1: B{Normal trust}. Run with the same privileges as a normal
                  user, that is, one that doesn't have the I{Administrator} or
                  I{Power User} user rights. Only available since Windows XP
                  and 2003, desktop editions.
             - 2: B{Full trust}. Run with the exact same privileges as the
                  current user. This is the default in normal mode.

        @type    bAllowElevation: bool
        @keyword bAllowElevation: C{True} to allow the child process to keep
            UAC elevation, if the debugger itself is running elevated. C{False}
            to ensure the child process doesn't run with elevation. Defaults to
            C{True} in normal mode and C{False} in hostile mode.

            This flag is only meaningful on Windows Vista and above, and if the
            debugger itself is running with elevation. It can be used to make
            sure the child processes don't run elevated as well.

            This flag DOES NOT force an elevation prompt when the debugger is
            not running with elevation.

            Note that running the debugger with elevation (or the Python
            interpreter at all for that matter) is not normally required.
            You should only need to if the target program requires elevation
            to work properly (for example if you try to debug an installer).

        @rtype:  L{Process}
        @return: A new Process object. Normally you don't need to use it now,
            it's best to interact with the process from the event handler.

        @raise WindowsError: Raises an exception on error.
        """
        if type(lpCmdLine) not in (str, compat.unicode):
            warnings.warn("Debug.execl expects a string")

        # Set the "debug" flag to True.
        kwargs["bDebug"] = True

        # Pop the "break on entry point" flag.
        bBreakOnEntryPoint = kwargs.pop("bBreakOnEntryPoint", False)

        # Set the default trust level if requested.
        if "iTrustLevel" not in kwargs:
            if self.__bHostileCode:
                kwargs["iTrustLevel"] = 0
            else:
                kwargs["iTrustLevel"] = 2

        # Set the default UAC elevation flag if requested.
        if "bAllowElevation" not in kwargs:
            kwargs["bAllowElevation"] = not self.__bHostileCode

        # In hostile mode the default parent process is explorer.exe.
        # Only supported for Windows Vista and above.
        if self.__bHostileCode and not kwargs.get("dwParentProcessId", None):
            try:
                vista_and_above = self.__vista_and_above
            except AttributeError:
                osi = win32.OSVERSIONINFOEXW()
                osi.dwMajorVersion = 6
                osi.dwMinorVersion = 0
                osi.dwPlatformId = win32.VER_PLATFORM_WIN32_NT
                mask = 0
                mask = win32.VerSetConditionMask(mask, win32.VER_MAJORVERSION, win32.VER_GREATER_EQUAL)
                mask = win32.VerSetConditionMask(mask, win32.VER_MAJORVERSION, win32.VER_GREATER_EQUAL)
                mask = win32.VerSetConditionMask(mask, win32.VER_PLATFORMID, win32.VER_EQUAL)
                vista_and_above = win32.VerifyVersionInfoW(
                    osi, win32.VER_MAJORVERSION | win32.VER_MINORVERSION | win32.VER_PLATFORMID, mask
                )
                self.__vista_and_above = vista_and_above
            if vista_and_above:
                dwParentProcessId = self.system.get_explorer_pid()
                if dwParentProcessId:
                    kwargs["dwParentProcessId"] = dwParentProcessId
                else:
                    msg = 'Failed to find "explorer.exe"!' " Using the debugger as parent process."
                    warnings.warn(msg, RuntimeWarning)

        # Start the new process.
        aProcess = None
        try:
            aProcess = self.system.start_process(lpCmdLine, **kwargs)
            dwProcessId = aProcess.get_pid()

            # Match the system kill-on-exit flag to our own.
            self.__setSystemKillOnExitMode()

            # Warn when mixing 32 and 64 bits.
            # This also allows the user to stop attaching altogether,
            # depending on how the warnings are configured.
            if System.bits != aProcess.get_bits():
                msg = "Mixture of 32 and 64 bits is considered experimental." " Use at your own risk!"
                warnings.warn(msg, MixedBitsWarning)

            # Add the new PID to the set of debugees.
            self.__startedDebugees.add(dwProcessId)

            # Add the new PID to the set of "break on EP" debugees if needed.
            if bBreakOnEntryPoint:
                self.__breakOnEP.add(dwProcessId)

            # Return the Process object.
            return aProcess

        # On error kill the new process and raise an exception.
        except:
            if aProcess is not None:
                try:
                    try:
                        self.__startedDebugees.remove(aProcess.get_pid())
                    except KeyError:
                        pass
                finally:
                    try:
                        try:
                            self.__breakOnEP.remove(aProcess.get_pid())
                        except KeyError:
                            pass
                    finally:
                        try:
                            aProcess.kill()
                        except Exception:
                            pass
            raise

    def add_existing_session(self, dwProcessId, bStarted=False):
        """
        Use this method only when for some reason the debugger's been attached
        to the target outside of WinAppDbg (for example when integrating with
        other tools).

        You don't normally need to call this method. Most users should call
        L{attach}, L{execv} or L{execl} instead.

        @type  dwProcessId: int
        @param dwProcessId: Global process ID.

        @type  bStarted: bool
        @param bStarted: C{True} if the process was started by the debugger,
            or C{False} if the process was attached to instead.

        @raise WindowsError: The target process does not exist, is not attached
            to the debugger anymore.
        """

        # Register the process object with the snapshot.
        if not self.system.has_process(dwProcessId):
            aProcess = Process(dwProcessId)
            self.system._add_process(aProcess)
        else:
            aProcess = self.system.get_process(dwProcessId)

        # Test for debug privileges on the target process.
        # Raises WindowsException on error.
        aProcess.get_handle()

        # Register the process ID with the debugger.
        if bStarted:
            self.__attachedDebugees.add(dwProcessId)
        else:
            self.__startedDebugees.add(dwProcessId)

        # Match the system kill-on-exit flag to our own.
        self.__setSystemKillOnExitMode()

        # Scan the process threads and loaded modules.
        # This is prefered because the thread and library events do not
        # properly give some information, like the filename for each module.
        aProcess.scan_threads()
        aProcess.scan_modules()

    def __cleanup_process(self, dwProcessId, bIgnoreExceptions=False):
        """
        Perform the necessary cleanup of a process about to be killed or
        detached from.

        This private method is called by L{kill} and L{detach}.

        @type  dwProcessId: int
        @param dwProcessId: Global ID of a process to kill.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when killing the process.

        @raise WindowsError: Raises an exception on error, unless
            C{bIgnoreExceptions} is C{True}.
        """
        # If the process is being debugged...
        if self.is_debugee(dwProcessId):
            # Make sure a Process object exists or the following calls fail.
            if not self.system.has_process(dwProcessId):
                aProcess = Process(dwProcessId)
                try:
                    aProcess.get_handle()
                except WindowsError:
                    pass  # fails later on with more specific reason
                self.system._add_process(aProcess)

            # Erase all breakpoints in the process.
            try:
                self.erase_process_breakpoints(dwProcessId)
            except Exception:
                if not bIgnoreExceptions:
                    raise
                e = sys.exc_info()[1]
                warnings.warn(str(e), RuntimeWarning)

            # Stop tracing all threads in the process.
            try:
                self.stop_tracing_process(dwProcessId)
            except Exception:
                if not bIgnoreExceptions:
                    raise
                e = sys.exc_info()[1]
                warnings.warn(str(e), RuntimeWarning)

            # The process is no longer a debugee.
            try:
                if dwProcessId in self.__attachedDebugees:
                    self.__attachedDebugees.remove(dwProcessId)
                if dwProcessId in self.__startedDebugees:
                    self.__startedDebugees.remove(dwProcessId)
            except Exception:
                if not bIgnoreExceptions:
                    raise
                e = sys.exc_info()[1]
                warnings.warn(str(e), RuntimeWarning)

        # Clear and remove the process from the snapshot.
        # If the user wants to do something with it after detaching
        # a new Process instance should be created.
        try:
            if self.system.has_process(dwProcessId):
                try:
                    self.system.get_process(dwProcessId).clear()
                finally:
                    self.system._del_process(dwProcessId)
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

        # If the last debugging event is related to this process, forget it.
        try:
            if self.lastEvent and self.lastEvent.get_pid() == dwProcessId:
                self.lastEvent = None
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

    def kill(self, dwProcessId, bIgnoreExceptions=False):
        """
        Kills a process currently being debugged.

        @see: L{detach}

        @type  dwProcessId: int
        @param dwProcessId: Global ID of a process to kill.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when killing the process.

        @raise WindowsError: Raises an exception on error, unless
            C{bIgnoreExceptions} is C{True}.
        """

        # Keep a reference to the process. We'll need it later.
        try:
            aProcess = self.system.get_process(dwProcessId)
        except KeyError:
            aProcess = Process(dwProcessId)

        # Cleanup all data referring to the process.
        self.__cleanup_process(dwProcessId, bIgnoreExceptions=bIgnoreExceptions)

        # Kill the process.
        try:
            try:
                if self.is_debugee(dwProcessId):
                    try:
                        if aProcess.is_alive():
                            aProcess.suspend()
                    finally:
                        self.detach(dwProcessId, bIgnoreExceptions=bIgnoreExceptions)
            finally:
                aProcess.kill()
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

        # Cleanup what remains of the process data.
        try:
            aProcess.clear()
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

    def kill_all(self, bIgnoreExceptions=False):
        """
        Kills from all processes currently being debugged.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when killing each process. C{False} to stop and raise an
            exception when encountering an error.

        @raise WindowsError: Raises an exception on error, unless
            C{bIgnoreExceptions} is C{True}.
        """
        for pid in self.get_debugee_pids():
            self.kill(pid, bIgnoreExceptions=bIgnoreExceptions)

    def detach(self, dwProcessId, bIgnoreExceptions=False):
        """
        Detaches from a process currently being debugged.

        @note: On Windows 2000 and below the process is killed.

        @see: L{attach}, L{detach_from_all}

        @type  dwProcessId: int
        @param dwProcessId: Global ID of a process to detach from.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when detaching. C{False} to stop and raise an exception when
            encountering an error.

        @raise WindowsError: Raises an exception on error, unless
            C{bIgnoreExceptions} is C{True}.
        """

        # Keep a reference to the process. We'll need it later.
        try:
            aProcess = self.system.get_process(dwProcessId)
        except KeyError:
            aProcess = Process(dwProcessId)

        # Determine if there is support for detaching.
        # This check should only fail on Windows 2000 and older.
        try:
            win32.DebugActiveProcessStop
            can_detach = True
        except AttributeError:
            can_detach = False

        # Continue the last event before detaching.
        # XXX not sure about this...
        try:
            if can_detach and self.lastEvent and self.lastEvent.get_pid() == dwProcessId:
                self.cont(self.lastEvent)
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

        # Cleanup all data referring to the process.
        self.__cleanup_process(dwProcessId, bIgnoreExceptions=bIgnoreExceptions)

        try:
            # Detach from the process.
            # On Windows 2000 and before, kill the process.
            if can_detach:
                try:
                    win32.DebugActiveProcessStop(dwProcessId)
                except Exception:
                    if not bIgnoreExceptions:
                        raise
                    e = sys.exc_info()[1]
                    warnings.warn(str(e), RuntimeWarning)
            else:
                try:
                    aProcess.kill()
                except Exception:
                    if not bIgnoreExceptions:
                        raise
                    e = sys.exc_info()[1]
                    warnings.warn(str(e), RuntimeWarning)

        finally:
            # Cleanup what remains of the process data.
            aProcess.clear()

    def detach_from_all(self, bIgnoreExceptions=False):
        """
        Detaches from all processes currently being debugged.

        @note: To better handle last debugging event, call L{stop} instead.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when detaching.

        @raise WindowsError: Raises an exception on error, unless
            C{bIgnoreExceptions} is C{True}.
        """
        for pid in self.get_debugee_pids():
            self.detach(pid, bIgnoreExceptions=bIgnoreExceptions)

    # ------------------------------------------------------------------------------

    def wait(self, dwMilliseconds=None):
        """
        Waits for the next debug event.

        @see: L{cont}, L{dispatch}, L{loop}

        @type  dwMilliseconds: int
        @param dwMilliseconds: (Optional) Timeout in milliseconds.
            Use C{INFINITE} or C{None} for no timeout.

        @rtype:  L{Event}
        @return: An event that occured in one of the debugees.

        @raise WindowsError: Raises an exception on error.
            If no target processes are left to debug,
            the error code is L{win32.ERROR_INVALID_HANDLE}.
        """

        # Wait for the next debug event.
        raw = win32.WaitForDebugEvent(dwMilliseconds)
        event = EventFactory.get(self, raw)

        # Remember it.
        self.lastEvent = event

        # Return it.
        return event

    def dispatch(self, event=None):
        """
        Calls the debug event notify callbacks.

        @see: L{cont}, L{loop}, L{wait}

        @type  event: L{Event}
        @param event: (Optional) Event object returned by L{wait}.

        @raise WindowsError: Raises an exception on error.
        """

        # If no event object was given, use the last event.
        if event is None:
            event = self.lastEvent

        # Ignore dummy events.
        if not event:
            return

        # Determine the default behaviour for this event.
        # XXX HACK
        # Some undocumented flags are used, but as far as I know in those
        # versions of Windows that don't support them they should behave
        # like DGB_CONTINUE.

        code = event.get_event_code()
        if code == win32.EXCEPTION_DEBUG_EVENT:
            # At this point, by default some exception types are swallowed by
            # the debugger, because we don't know yet if it was caused by the
            # debugger itself or the debugged process.
            #
            # Later on (see breakpoint.py) if we determined the exception was
            # not caused directly by the debugger itself, we set the default
            # back to passing the exception to the debugee.
            #
            # The "invalid handle" exception is also swallowed by the debugger
            # because it's not normally generated by the debugee. But in
            # hostile mode we want to pass it to the debugee, as it may be the
            # result of an anti-debug trick. In that case it's best to disable
            # bad handles detection with Microsoft's gflags.exe utility. See:
            # http://msdn.microsoft.com/en-us/library/windows/hardware/ff549557(v=vs.85).aspx

            exc_code = event.get_exception_code()
            if exc_code in (
                win32.EXCEPTION_BREAKPOINT,
                win32.EXCEPTION_WX86_BREAKPOINT,
                win32.EXCEPTION_SINGLE_STEP,
                win32.EXCEPTION_GUARD_PAGE,
            ):
                event.continueStatus = win32.DBG_CONTINUE
            elif exc_code == win32.EXCEPTION_INVALID_HANDLE:
                if self.__bHostileCode:
                    event.continueStatus = win32.DBG_EXCEPTION_NOT_HANDLED
                else:
                    event.continueStatus = win32.DBG_CONTINUE
            else:
                event.continueStatus = win32.DBG_EXCEPTION_NOT_HANDLED

        elif code == win32.RIP_EVENT and event.get_rip_type() == win32.SLE_ERROR:
            # RIP events that signal fatal events should kill the process.
            event.continueStatus = win32.DBG_TERMINATE_PROCESS

        else:
            # Other events need this continue code.
            # Sometimes other codes can be used and are ignored, sometimes not.
            # For example, when using the DBG_EXCEPTION_NOT_HANDLED code,
            # debug strings are sent twice (!)
            event.continueStatus = win32.DBG_CONTINUE

        # Dispatch the debug event.
        return EventDispatcher.dispatch(self, event)

    def cont(self, event=None):
        """
        Resumes execution after processing a debug event.

        @see: dispatch(), loop(), wait()

        @type  event: L{Event}
        @param event: (Optional) Event object returned by L{wait}.

        @raise WindowsError: Raises an exception on error.
        """

        # If no event object was given, use the last event.
        if event is None:
            event = self.lastEvent

        # Ignore dummy events.
        if not event:
            return

        # Get the event continue status information.
        dwProcessId = event.get_pid()
        dwThreadId = event.get_tid()
        dwContinueStatus = event.continueStatus

        # Check if the process is still being debugged.
        if self.is_debugee(dwProcessId):
            # Try to flush the instruction cache.
            try:
                if self.system.has_process(dwProcessId):
                    aProcess = self.system.get_process(dwProcessId)
                else:
                    aProcess = Process(dwProcessId)
                aProcess.flush_instruction_cache()
            except WindowsError:
                pass

            # XXX TODO
            #
            # Try to execute the UnhandledExceptionFilter for second chance
            # exceptions, at least when in hostile mode (in normal mode it
            # would be breaking compatibility, as users may actually expect
            # second chance exceptions to be raised again).
            #
            # Reportedly in Windows 7 (maybe in Vista too) this seems to be
            # happening already. In XP and below the UnhandledExceptionFilter
            # was never called for processes being debugged.

            # Continue execution of the debugee.
            win32.ContinueDebugEvent(dwProcessId, dwThreadId, dwContinueStatus)

        # If the event is the last event, forget it.
        if event == self.lastEvent:
            self.lastEvent = None

    def stop(self, bIgnoreExceptions=True):
        """
        Stops debugging all processes.

        If the kill on exit mode is on, debugged processes are killed when the
        debugger is stopped. Otherwise when the debugger stops it detaches from
        all debugged processes and leaves them running (default). For more
        details see: L{__init__}

        @note: This method is better than L{detach_from_all} because it can
            gracefully handle the last debugging event before detaching.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when detaching.
        """

        # Determine if we have a last debug event that we need to continue.
        try:
            event = self.lastEvent
            has_event = bool(event)
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)
            has_event = False

        # If we do...
        if has_event:
            # Disable all breakpoints in the process before resuming execution.
            try:
                pid = event.get_pid()
                self.disable_process_breakpoints(pid)
            except Exception:
                if not bIgnoreExceptions:
                    raise
                e = sys.exc_info()[1]
                warnings.warn(str(e), RuntimeWarning)

            # Disable all breakpoints in the thread before resuming execution.
            try:
                tid = event.get_tid()
                self.disable_thread_breakpoints(tid)
            except Exception:
                if not bIgnoreExceptions:
                    raise
                e = sys.exc_info()[1]
                warnings.warn(str(e), RuntimeWarning)

            # Resume execution.
            try:
                event.continueDebugEvent = win32.DBG_CONTINUE
                self.cont(event)
            except Exception:
                if not bIgnoreExceptions:
                    raise
                e = sys.exc_info()[1]
                warnings.warn(str(e), RuntimeWarning)

        # Detach from or kill all debuggees.
        try:
            if self.__bKillOnExit:
                self.kill_all(bIgnoreExceptions)
            else:
                self.detach_from_all(bIgnoreExceptions)
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

        # Cleanup the process snapshots.
        try:
            self.system.clear()
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

        # Close all Win32 handles the Python garbage collector failed to close.
        self.force_garbage_collection(bIgnoreExceptions)

    def next(self):
        """
        Handles the next debug event.

        @see: L{cont}, L{dispatch}, L{wait}, L{stop}

        @raise WindowsError: Raises an exception on error.

            If the wait operation causes an error, debugging is stopped
            (meaning all debugees are either killed or detached from).

            If the event dispatching causes an error, the event is still
            continued before returning. This may happen, for example, if the
            event handler raises an exception nobody catches.
        """
        try:
            event = self.wait()
        except Exception:
            self.stop()
            raise
        try:
            self.dispatch()
        finally:
            self.cont()

    def loop(self):
        """
        Simple debugging loop.

        This debugging loop is meant to be useful for most simple scripts.
        It iterates as long as there is at least one debugee, or an exception
        is raised. Multiple calls are allowed.

        This is a trivial example script::
            import sys
            debug = Debug()
            try:
                debug.execv( sys.argv [ 1 : ] )
                debug.loop()
            finally:
                debug.stop()

        @see: L{next}, L{stop}

            U{http://msdn.microsoft.com/en-us/library/ms681675(VS.85).aspx}

        @raise WindowsError: Raises an exception on error.

            If the wait operation causes an error, debugging is stopped
            (meaning all debugees are either killed or detached from).

            If the event dispatching causes an error, the event is still
            continued before returning. This may happen, for example, if the
            event handler raises an exception nobody catches.
        """
        while self:
            self.next()

    def get_debugee_count(self):
        """
        @rtype:  int
        @return: Number of processes being debugged.
        """
        return len(self.__attachedDebugees) + len(self.__startedDebugees)

    def get_debugee_pids(self):
        """
        @rtype:  list( int... )
        @return: Global IDs of processes being debugged.
        """
        return list(self.__attachedDebugees) + list(self.__startedDebugees)

    def is_debugee(self, dwProcessId):
        """
        Determine if the debugger is debugging the given process.

        @see: L{is_debugee_attached}, L{is_debugee_started}

        @type  dwProcessId: int
        @param dwProcessId: Process global ID.

        @rtype:  bool
        @return: C{True} if the given process is being debugged
            by this L{Debug} instance.
        """
        return self.is_debugee_attached(dwProcessId) or self.is_debugee_started(dwProcessId)

    def is_debugee_started(self, dwProcessId):
        """
        Determine if the given process was started by the debugger.

        @see: L{is_debugee}, L{is_debugee_attached}

        @type  dwProcessId: int
        @param dwProcessId: Process global ID.

        @rtype:  bool
        @return: C{True} if the given process was started for debugging by this
            L{Debug} instance.
        """
        return dwProcessId in self.__startedDebugees

    def is_debugee_attached(self, dwProcessId):
        """
        Determine if the debugger is attached to the given process.

        @see: L{is_debugee}, L{is_debugee_started}

        @type  dwProcessId: int
        @param dwProcessId: Process global ID.

        @rtype:  bool
        @return: C{True} if the given process is attached to this
            L{Debug} instance.
        """
        return dwProcessId in self.__attachedDebugees

    def in_hostile_mode(self):
        """
        Determine if we're in hostile mode (anti-anti-debug).

        @rtype:  bool
        @return: C{True} if this C{Debug} instance was started in hostile mode,
            C{False} otherwise.
        """
        return self.__bHostileCode

    # ------------------------------------------------------------------------------

    def interactive(self, bConfirmQuit=True, bShowBanner=True):
        """
        Start an interactive debugging session.

        @type  bConfirmQuit: bool
        @param bConfirmQuit: Set to C{True} to ask the user for confirmation
            before closing the session, C{False} otherwise.

        @type  bShowBanner: bool
        @param bShowBanner: Set to C{True} to show a banner before entering
            the session and after leaving it, C{False} otherwise.

        @warn: This will temporarily disable the user-defined event handler!

        This method returns when the user closes the session.
        """
        print("")
        print("-" * 79)
        print("Interactive debugging session started.")
        print('Use the "help" command to list all available commands.')
        print('Use the "quit" command to close this session.')
        print("-" * 79)
        if self.lastEvent is None:
            print("")
        console = ConsoleDebugger()
        console.confirm_quit = bConfirmQuit
        console.load_history()
        try:
            console.start_using_debugger(self)
            console.loop()
        finally:
            console.stop_using_debugger()
            console.save_history()
        print("")
        print("-" * 79)
        print("Interactive debugging session closed.")
        print("-" * 79)
        print("")

    # ------------------------------------------------------------------------------

    @staticmethod
    def force_garbage_collection(bIgnoreExceptions=True):
        """
        Close all Win32 handles the Python garbage collector failed to close.

        @type  bIgnoreExceptions: bool
        @param bIgnoreExceptions: C{True} to ignore any exceptions that may be
            raised when detaching.
        """
        try:
            import gc

            gc.collect()
            bRecollect = False
            for obj in list(gc.garbage):
                try:
                    if isinstance(obj, win32.Handle):
                        obj.close()
                    elif isinstance(obj, Event):
                        obj.debug = None
                    elif isinstance(obj, Process):
                        obj.clear()
                    elif isinstance(obj, Thread):
                        obj.set_process(None)
                        obj.clear()
                    elif isinstance(obj, Module):
                        obj.set_process(None)
                    elif isinstance(obj, Window):
                        obj.set_process(None)
                    else:
                        continue
                    gc.garbage.remove(obj)
                    del obj
                    bRecollect = True
                except Exception:
                    if not bIgnoreExceptions:
                        raise
                    e = sys.exc_info()[1]
                    warnings.warn(str(e), RuntimeWarning)
            if bRecollect:
                gc.collect()
        except Exception:
            if not bIgnoreExceptions:
                raise
            e = sys.exc_info()[1]
            warnings.warn(str(e), RuntimeWarning)

    # ------------------------------------------------------------------------------

    def _notify_create_process(self, event):
        """
        Notify the creation of a new process.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{CreateProcessEvent}
        @param event: Create process event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        dwProcessId = event.get_pid()
        if dwProcessId not in self.__attachedDebugees:
            if dwProcessId not in self.__startedDebugees:
                self.__startedDebugees.add(dwProcessId)

        retval = self.system._notify_create_process(event)

        # Set a breakpoint on the program's entry point if requested.
        # Try not to use the Event object's entry point value, as in some cases
        # it may be wrong. See: http://pferrie.host22.com/misc/lowlevel3.htm
        if dwProcessId in self.__breakOnEP:
            try:
                lpEntryPoint = event.get_process().get_entry_point()
            except Exception:
                lpEntryPoint = event.get_start_address()

            # It'd be best to use a hardware breakpoint instead, at least in
            # hostile mode. But since the main thread's context gets smashed
            # by the loader, I haven't found a way to make it work yet.
            self.break_at(dwProcessId, lpEntryPoint)

        # Defeat isDebuggerPresent by patching PEB->BeingDebugged.
        # When we do this, some debugging APIs cease to work as expected.
        # For example, the system breakpoint isn't hit when we attach.
        # For that reason we need to define a code breakpoint at the
        # code location where a new thread is spawned by the debugging
        # APIs, ntdll!DbgUiRemoteBreakin.
        if self.__bHostileCode:
            aProcess = event.get_process()
            try:
                hProcess = aProcess.get_handle(win32.PROCESS_QUERY_INFORMATION)
                pbi = win32.NtQueryInformationProcess(hProcess, win32.ProcessBasicInformation)
                ptr = pbi.PebBaseAddress + 2
                if aProcess.peek(ptr, 1) == "\x01":
                    aProcess.poke(ptr, "\x00")
            except WindowsError:
                e = sys.exc_info()[1]
                warnings.warn("Cannot patch PEB->BeingDebugged, reason: %s" % e.strerror)

        return retval

    def _notify_create_thread(self, event):
        """
        Notify the creation of a new thread.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{CreateThreadEvent}
        @param event: Create thread event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        return event.get_process()._notify_create_thread(event)

    def _notify_load_dll(self, event):
        """
        Notify the load of a new module.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{LoadDLLEvent}
        @param event: Load DLL event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """

        # Pass the event to the breakpoint container.
        bCallHandler = _BreakpointContainer._notify_load_dll(self, event)

        # Get the process where the DLL was loaded.
        aProcess = event.get_process()

        # Pass the event to the process.
        bCallHandler = aProcess._notify_load_dll(event) and bCallHandler

        # Anti-anti-debugging tricks on ntdll.dll.
        if self.__bHostileCode:
            aModule = event.get_module()
            if aModule.match_name("ntdll.dll"):
                # Since we've overwritten the PEB to hide
                # ourselves, we no longer have the system
                # breakpoint when attaching to the process.
                # Set a breakpoint at ntdll!DbgUiRemoteBreakin
                # instead (that's where the debug API spawns
                # it's auxiliary threads). This also defeats
                # a simple anti-debugging trick: the hostile
                # process could have overwritten the int3
                # instruction at the system breakpoint.
                self.break_at(aProcess.get_pid(), aProcess.resolve_label("ntdll!DbgUiRemoteBreakin"))

        return bCallHandler

    def _notify_exit_process(self, event):
        """
        Notify the termination of a process.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{ExitProcessEvent}
        @param event: Exit process event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        bCallHandler1 = _BreakpointContainer._notify_exit_process(self, event)
        bCallHandler2 = self.system._notify_exit_process(event)

        try:
            self.detach(event.get_pid())
        except WindowsError:
            e = sys.exc_info()[1]
            if e.winerror != win32.ERROR_INVALID_PARAMETER:
                warnings.warn("Failed to detach from dead process, reason: %s" % str(e), RuntimeWarning)
        except Exception:
            e = sys.exc_info()[1]
            warnings.warn("Failed to detach from dead process, reason: %s" % str(e), RuntimeWarning)

        return bCallHandler1 and bCallHandler2

    def _notify_exit_thread(self, event):
        """
        Notify the termination of a thread.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{ExitThreadEvent}
        @param event: Exit thread event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        bCallHandler1 = _BreakpointContainer._notify_exit_thread(self, event)
        bCallHandler2 = event.get_process()._notify_exit_thread(event)
        return bCallHandler1 and bCallHandler2

    def _notify_unload_dll(self, event):
        """
        Notify the unload of a module.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{UnloadDLLEvent}
        @param event: Unload DLL event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        bCallHandler1 = _BreakpointContainer._notify_unload_dll(self, event)
        bCallHandler2 = event.get_process()._notify_unload_dll(event)
        return bCallHandler1 and bCallHandler2

    def _notify_rip(self, event):
        """
        Notify of a RIP event.

        @warning: This method is meant to be used internally by the debugger.

        @type  event: L{RIPEvent}
        @param event: RIP event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        event.debug.detach(event.get_pid())
        return True

    def _notify_debug_control_c(self, event):
        """
        Notify of a Debug Ctrl-C exception.

        @warning: This method is meant to be used internally by the debugger.

        @note: This exception is only raised when a debugger is attached, and
            applications are not supposed to handle it, so we need to handle it
            ourselves or the application may crash.

        @see: U{http://msdn.microsoft.com/en-us/library/aa363082(VS.85).aspx}

        @type  event: L{ExceptionEvent}
        @param event: Debug Ctrl-C exception event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        if event.is_first_chance():
            event.continueStatus = win32.DBG_EXCEPTION_HANDLED
        return True

    def _notify_ms_vc_exception(self, event):
        """
        Notify of a Microsoft Visual C exception.

        @warning: This method is meant to be used internally by the debugger.

        @note: This allows the debugger to understand the
            Microsoft Visual C thread naming convention.

        @see: U{http://msdn.microsoft.com/en-us/library/xcb2z8hs.aspx}

        @type  event: L{ExceptionEvent}
        @param event: Microsoft Visual C exception event.

        @rtype:  bool
        @return: C{True} to call the user-defined handle, C{False} otherwise.
        """
        dwType = event.get_exception_information(0)
        if dwType == 0x1000:
            pszName = event.get_exception_information(1)
            dwThreadId = event.get_exception_information(2)
            dwFlags = event.get_exception_information(3)

            aProcess = event.get_process()
            szName = aProcess.peek_string(pszName, fUnicode=False)
            if szName:
                if dwThreadId == -1:
                    dwThreadId = event.get_tid()

                if aProcess.has_thread(dwThreadId):
                    aThread = aProcess.get_thread(dwThreadId)
                else:
                    aThread = Thread(dwThreadId)
                    aProcess._add_thread(aThread)

                ##                if aThread.get_name() is None:
                ##                    aThread.set_name(szName)
                aThread.set_name(szName)

        return True



from _pydevd_bundle.pydevd_constants import STATE_SUSPEND, JINJA2_SUSPEND
from _pydevd_bundle.pydevd_comm import CMD_SET_BREAK, CMD_ADD_EXCEPTION_BREAK
from pydevd_file_utils import canonical_normalized_path
from _pydevd_bundle.pydevd_frame_utils import add_exception_to_frame, FCode
from _pydev_bundle import pydev_log
from pydevd_plugins.pydevd_line_validation import LineBreakpointWithLazyValidation, ValidationInfo
from _pydev_bundle.pydev_override import overrides
from _pydevd_bundle.pydevd_api import PyDevdAPI


class Jinja2LineBreakpoint(LineBreakpointWithLazyValidation):
    def __init__(
        self, canonical_normalized_filename, breakpoint_id, line, condition, func_name, expression, hit_condition=None, is_logpoint=False
    ):
        self.canonical_normalized_filename = canonical_normalized_filename
        LineBreakpointWithLazyValidation.__init__(
            self, breakpoint_id, line, condition, func_name, expression, hit_condition=hit_condition, is_logpoint=is_logpoint
        )

    def __str__(self):
        return "Jinja2LineBreakpoint: %s-%d" % (self.canonical_normalized_filename, self.line)


class _Jinja2ValidationInfo(ValidationInfo):
    @overrides(ValidationInfo._collect_valid_lines_in_template_uncached)
    def _collect_valid_lines_in_template_uncached(self, template):
        lineno_mapping = _get_frame_lineno_mapping(template)
        if not lineno_mapping:
            return set()

        return set(x[0] for x in lineno_mapping)


def add_line_breakpoint(
    pydb,
    type,
    canonical_normalized_filename,
    breakpoint_id,
    line,
    condition,
    expression,
    func_name,
    hit_condition=None,
    is_logpoint=False,
    add_breakpoint_result=None,
    on_changed_breakpoint_state=None,
):
    if type == "jinja2-line":
        jinja2_line_breakpoint = Jinja2LineBreakpoint(
            canonical_normalized_filename,
            breakpoint_id,
            line,
            condition,
            func_name,
            expression,
            hit_condition=hit_condition,
            is_logpoint=is_logpoint,
        )
        if not hasattr(pydb, "jinja2_breakpoints"):
            _init_plugin_breaks(pydb)

        add_breakpoint_result.error_code = PyDevdAPI.ADD_BREAKPOINT_LAZY_VALIDATION
        jinja2_line_breakpoint.add_breakpoint_result = add_breakpoint_result
        jinja2_line_breakpoint.on_changed_breakpoint_state = on_changed_breakpoint_state

        return jinja2_line_breakpoint, pydb.jinja2_breakpoints
    return None


def after_breakpoints_consolidated(py_db, canonical_normalized_filename, id_to_pybreakpoint, file_to_line_to_breakpoints):
    jinja2_breakpoints_for_file = file_to_line_to_breakpoints.get(canonical_normalized_filename)
    if not jinja2_breakpoints_for_file:
        return

    if not hasattr(py_db, "jinja2_validation_info"):
        _init_plugin_breaks(py_db)

    # In general we validate the breakpoints only when the template is loaded, but if the template
    # was already loaded, we can validate the breakpoints based on the last loaded value.
    py_db.jinja2_validation_info.verify_breakpoints_from_template_cached_lines(
        py_db, canonical_normalized_filename, jinja2_breakpoints_for_file
    )


def add_exception_breakpoint(pydb, type, exception):
    if type == "jinja2":
        if not hasattr(pydb, "jinja2_exception_break"):
            _init_plugin_breaks(pydb)
        pydb.jinja2_exception_break[exception] = True
        return True
    return False


def _init_plugin_breaks(pydb):
    pydb.jinja2_exception_break = {}
    pydb.jinja2_breakpoints = {}

    pydb.jinja2_validation_info = _Jinja2ValidationInfo()


def remove_all_exception_breakpoints(pydb):
    if hasattr(pydb, "jinja2_exception_break"):
        pydb.jinja2_exception_break = {}
        return True
    return False


def remove_exception_breakpoint(pydb, exception_type, exception):
    if exception_type == "jinja2":
        try:
            del pydb.jinja2_exception_break[exception]
            return True
        except KeyError:
            pass
    return False


def get_breakpoints(pydb, breakpoint_type):
    if breakpoint_type == "jinja2-line":
        return pydb.jinja2_breakpoints
    return None


def _is_jinja2_render_call(frame):
    try:
        name = frame.f_code.co_name
        if "__jinja_template__" in frame.f_globals and name in ("root", "loop", "macro") or name.startswith("block_"):
            return True
        return False
    except:
        pydev_log.exception()
        return False


def _suspend_jinja2(pydb, thread, frame, cmd=CMD_SET_BREAK, message=None):
    frame = Jinja2TemplateFrame(frame)

    if frame.f_lineno is None:
        return None

    pydb.set_suspend(thread, cmd)

    thread.additional_info.suspend_type = JINJA2_SUSPEND
    if cmd == CMD_ADD_EXCEPTION_BREAK:
        # send exception name as message
        if message:
            message = str(message)
        thread.additional_info.pydev_message = message

    return frame


def _is_jinja2_suspended(thread):
    return thread.additional_info.suspend_type == JINJA2_SUSPEND


def _is_jinja2_context_call(frame):
    return "_Context__obj" in frame.f_locals


def _is_jinja2_internal_function(frame):
    return "self" in frame.f_locals and frame.f_locals["self"].__class__.__name__ in (
        "LoopContext",
        "TemplateReference",
        "Macro",
        "BlockReference",
    )


def _find_jinja2_render_frame(frame):
    while frame is not None and not _is_jinja2_render_call(frame):
        frame = frame.f_back

    return frame


# =======================================================================================================================
# Jinja2 Frame
# =======================================================================================================================


class Jinja2TemplateFrame(object):
    IS_PLUGIN_FRAME = True

    def __init__(self, frame, original_filename=None, template_lineno=None):
        if original_filename is None:
            original_filename = _get_jinja2_template_original_filename(frame)

        if template_lineno is None:
            template_lineno = _get_jinja2_template_line(frame)

        self.back_context = None
        if "context" in frame.f_locals:
            # sometimes we don't have 'context', e.g. in macros
            self.back_context = frame.f_locals["context"]
        self.f_code = FCode("template", original_filename)
        self.f_lineno = template_lineno
        self.f_back = frame
        self.f_globals = {}
        self.f_locals = self.collect_context(frame)
        self.f_trace = None

    def _get_real_var_name(self, orig_name):
        # replace leading number for local variables
        parts = orig_name.split("_")
        if len(parts) > 1 and parts[0].isdigit():
            return parts[1]
        return orig_name

    def collect_context(self, frame):
        res = {}
        for k, v in frame.f_locals.items():
            if not k.startswith("l_"):
                res[k] = v
            elif v and not _is_missing(v):
                res[self._get_real_var_name(k[2:])] = v
        if self.back_context is not None:
            for k, v in self.back_context.items():
                res[k] = v
        return res

    def _change_variable(self, frame, name, value):
        in_vars_or_parents = False
        if "context" in frame.f_locals:
            if name in frame.f_locals["context"].parent:
                self.back_context.parent[name] = value
                in_vars_or_parents = True
            if name in frame.f_locals["context"].vars:
                self.back_context.vars[name] = value
                in_vars_or_parents = True

        l_name = "l_" + name
        if l_name in frame.f_locals:
            if in_vars_or_parents:
                frame.f_locals[l_name] = self.back_context.resolve(name)
            else:
                frame.f_locals[l_name] = value


class Jinja2TemplateSyntaxErrorFrame(object):
    IS_PLUGIN_FRAME = True

    def __init__(self, frame, exception_cls_name, filename, lineno, f_locals):
        self.f_code = FCode("Jinja2 %s" % (exception_cls_name,), filename)
        self.f_lineno = lineno
        self.f_back = frame
        self.f_globals = {}
        self.f_locals = f_locals
        self.f_trace = None


def change_variable(frame, attr, expression, default, scope=None):
    if isinstance(frame, Jinja2TemplateFrame):
        result = eval(expression, frame.f_globals, frame.f_locals)
        frame._change_variable(frame.f_back, attr, result, scope=scope)
        return result
    return default


def _is_missing(item):
    if item.__class__.__name__ == "MissingType":
        return True
    return False


def _find_render_function_frame(frame):
    # in order to hide internal rendering functions
    old_frame = frame
    try:
        while not (
            "self" in frame.f_locals and frame.f_locals["self"].__class__.__name__ == "Template" and frame.f_code.co_name == "render"
        ):
            frame = frame.f_back
            if frame is None:
                return old_frame
        return frame
    except:
        return old_frame


def _get_jinja2_template_debug_info(frame):
    frame_globals = frame.f_globals

    jinja_template = frame_globals.get("__jinja_template__")

    if jinja_template is None:
        return None

    return _get_frame_lineno_mapping(jinja_template)


def _get_frame_lineno_mapping(jinja_template):
    """
    :rtype: list(tuple(int,int))
    :return: list((original_line, line_in_frame))
    """
    # _debug_info is a string with the mapping from frame line to actual line
    # i.e.: "5=13&8=14"
    _debug_info = jinja_template._debug_info
    if not _debug_info:
        # Sometimes template contains only plain text.
        return None

    # debug_info is a list with the mapping from frame line to actual line
    # i.e.: [(5, 13), (8, 14)]
    return jinja_template.debug_info


def _get_jinja2_template_line(frame):
    debug_info = _get_jinja2_template_debug_info(frame)
    if debug_info is None:
        return None

    lineno = frame.f_lineno

    for pair in debug_info:
        if pair[1] == lineno:
            return pair[0]

    return None


def _convert_to_str(s):
    return s


def _get_jinja2_template_original_filename(frame):
    if "__jinja_template__" in frame.f_globals:
        return _convert_to_str(frame.f_globals["__jinja_template__"].filename)

    return None


# =======================================================================================================================
# Jinja2 Step Commands
# =======================================================================================================================


def has_exception_breaks(py_db):
    if len(py_db.jinja2_exception_break) > 0:
        return True
    return False


def has_line_breaks(py_db):
    for _canonical_normalized_filename, breakpoints in py_db.jinja2_breakpoints.items():
        if len(breakpoints) > 0:
            return True
    return False


def can_skip(pydb, frame):
    if pydb.jinja2_breakpoints and _is_jinja2_render_call(frame):
        filename = _get_jinja2_template_original_filename(frame)
        if filename is not None:
            canonical_normalized_filename = canonical_normalized_path(filename)
            jinja2_breakpoints_for_file = pydb.jinja2_breakpoints.get(canonical_normalized_filename)
            if jinja2_breakpoints_for_file:
                return False

    if pydb.jinja2_exception_break:
        name = frame.f_code.co_name

        # errors in compile time
        if name in ("template", "top-level template code", "<module>") or name.startswith("block "):
            f_back = frame.f_back
            module_name = ""
            if f_back is not None:
                module_name = f_back.f_globals.get("__name__", "")
            if module_name.startswith("jinja2."):
                return False

    return True


is_tracked_frame = _is_jinja2_render_call


def required_events_breakpoint():
    return ("line",)


def required_events_stepping():
    return ("call", "line", "return")


def cmd_step_into(pydb, frame, event, info, thread, stop_info, stop):
    plugin_stop = False
    if _is_jinja2_suspended(thread):
        plugin_stop = stop_info["jinja2_stop"] = event in ("call", "line") and _is_jinja2_render_call(frame)
        stop = False
        if info.pydev_call_from_jinja2 is not None:
            if _is_jinja2_internal_function(frame):
                # if internal Jinja2 function was called, we should continue debugging inside template
                info.pydev_call_from_jinja2 = None
            else:
                # we go into python code from Jinja2 rendering frame
                stop = True

        if event == "call" and _is_jinja2_context_call(frame.f_back):
            # we called function from context, the next step will be in function
            info.pydev_call_from_jinja2 = 1

    if event == "return" and _is_jinja2_context_call(frame.f_back):
        # we return from python code to Jinja2 rendering frame
        info.pydev_step_stop = info.pydev_call_from_jinja2
        info.pydev_call_from_jinja2 = None
        thread.additional_info.suspend_type = JINJA2_SUSPEND
        stop = False

        # print "info.pydev_call_from_jinja2", info.pydev_call_from_jinja2, "stop_info", stop_info, \
        #    "thread.additional_info.suspend_type", thread.additional_info.suspend_type
        # print "event", event, "farme.locals", frame.f_locals
    return stop, plugin_stop


def cmd_step_over(pydb, frame, event, info, thread, stop_info, stop):
    plugin_stop = False
    if _is_jinja2_suspended(thread):
        stop = False

        if info.pydev_call_inside_jinja2 is None:
            if _is_jinja2_render_call(frame):
                if event == "call":
                    info.pydev_call_inside_jinja2 = frame.f_back
                if event in ("line", "return"):
                    info.pydev_call_inside_jinja2 = frame
        else:
            if event == "line":
                if _is_jinja2_render_call(frame) and info.pydev_call_inside_jinja2 is frame:
                    plugin_stop = stop_info["jinja2_stop"] = True
            if event == "return":
                if frame is info.pydev_call_inside_jinja2 and "event" not in frame.f_back.f_locals:
                    info.pydev_call_inside_jinja2 = _find_jinja2_render_frame(frame.f_back)
        return stop, plugin_stop
    else:
        if event == "return" and _is_jinja2_context_call(frame.f_back):
            # we return from python code to Jinja2 rendering frame
            info.pydev_call_from_jinja2 = None
            info.pydev_call_inside_jinja2 = _find_jinja2_render_frame(frame)
            thread.additional_info.suspend_type = JINJA2_SUSPEND
            stop = False
            return stop, plugin_stop
    # print "info.pydev_call_from_jinja2", info.pydev_call_from_jinja2, "stop", stop, "jinja_stop", jinja2_stop, \
    #    "thread.additional_info.suspend_type", thread.additional_info.suspend_type
    # print "event", event, "info.pydev_call_inside_jinja2", info.pydev_call_inside_jinja2
    # print "frame", frame, "frame.f_back", frame.f_back, "step_stop", info.pydev_step_stop
    # print "is_context_call", _is_jinja2_context_call(frame)
    # print "render", _is_jinja2_render_call(frame)
    # print "-------------"
    return stop, plugin_stop


def stop(pydb, frame, event, thread, stop_info, arg, step_cmd):
    if "jinja2_stop" in stop_info and stop_info["jinja2_stop"]:
        frame = _suspend_jinja2(pydb, thread, frame, step_cmd)
        if frame:
            pydb.do_wait_suspend(thread, frame, event, arg)
            return True
    return False


def get_breakpoint(py_db, frame, event, info):
    break_type = "jinja2"

    if event == "line" and info.pydev_state != STATE_SUSPEND and py_db.jinja2_breakpoints and _is_jinja2_render_call(frame):
        jinja_template = frame.f_globals.get("__jinja_template__")
        if jinja_template is None:
            return None

        original_filename = _get_jinja2_template_original_filename(frame)
        if original_filename is not None:
            pydev_log.debug("Jinja2 is rendering a template: %s", original_filename)
            canonical_normalized_filename = canonical_normalized_path(original_filename)
            jinja2_breakpoints_for_file = py_db.jinja2_breakpoints.get(canonical_normalized_filename)

            if jinja2_breakpoints_for_file:
                jinja2_validation_info = py_db.jinja2_validation_info
                jinja2_validation_info.verify_breakpoints(py_db, canonical_normalized_filename, jinja2_breakpoints_for_file, jinja_template)

                template_lineno = _get_jinja2_template_line(frame)
                if template_lineno is not None:
                    jinja2_breakpoint = jinja2_breakpoints_for_file.get(template_lineno)
                    if jinja2_breakpoint is not None:
                        new_frame = Jinja2TemplateFrame(frame, original_filename, template_lineno)
                        return jinja2_breakpoint, new_frame, break_type

    return None


def suspend(pydb, thread, frame, bp_type):
    if bp_type == "jinja2":
        return _suspend_jinja2(pydb, thread, frame)
    return None


def exception_break(pydb, frame, thread, arg, is_unwind):
    exception, value, trace = arg
    if pydb.jinja2_exception_break and exception is not None:
        exception_type = list(pydb.jinja2_exception_break.keys())[0]
        if exception.__name__ in ("UndefinedError", "TemplateNotFound", "TemplatesNotFound"):
            # errors in rendering
            render_frame = _find_jinja2_render_frame(frame)
            if render_frame:
                suspend_frame = _suspend_jinja2(pydb, thread, render_frame, CMD_ADD_EXCEPTION_BREAK, message=exception_type)
                if suspend_frame:
                    add_exception_to_frame(suspend_frame, (exception, value, trace))
                    suspend_frame.f_back = frame
                    frame = suspend_frame
                    return True, frame

        elif exception.__name__ in ("TemplateSyntaxError", "TemplateAssertionError"):
            name = frame.f_code.co_name

            # errors in compile time
            if name in ("template", "top-level template code", "<module>") or name.startswith("block "):
                f_back = frame.f_back
                if f_back is not None:
                    module_name = f_back.f_globals.get("__name__", "")

                if module_name.startswith("jinja2."):
                    # Jinja2 translates exception info and creates fake frame on his own
                    pydb.set_suspend(thread, CMD_ADD_EXCEPTION_BREAK)
                    add_exception_to_frame(frame, (exception, value, trace))
                    thread.additional_info.suspend_type = JINJA2_SUSPEND
                    thread.additional_info.pydev_message = str(exception_type)
                    return True, frame
    return None




# This file was generated by 'versioneer.py' (0.23) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2026-01-27T16:58:21-0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "7ac3d1f2c65ccf6b2e62a762863093f0242179f0",
 "version": "1.8.20"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import sys
from _pydev_bundle import pydev_log

backends = {
    "tk": "TkAgg",
    "gtk": "GTKAgg",
    "wx": "WXAgg",
    "qt": "QtAgg",  # Auto-choose qt4/5
    "qt4": "Qt4Agg",
    "qt5": "Qt5Agg",
    "qt6": "Qt6Agg",
    "osx": "MacOSX",
}

lowercase_convert = {
    "tkagg": "TkAgg",
    "gtkagg": "GTKAgg",
    "wxagg": "WXAgg",
    "qtagg": "QtAgg",
    "qt4agg": "Qt4Agg",
    "qt5agg": "Qt5Agg",
    "qt6agg": "Qt6Agg",
    "macosx": "MacOSX",
    "gtk": "GTK",
    "gtkcairo": "GTKCairo",
    "wx": "WX",
    "cocoaagg": "CocoaAgg",
}

# We also need a reverse backends2guis mapping that will properly choose which
# GUI support to activate based on the desired matplotlib backend.  For the
# most part it's just a reverse of the above dict, but we also need to add a
# few others that map to the same GUI manually:
backend2gui = dict(zip(backends.values(), backends.keys()))
# In the reverse mapping, there are a few extra valid matplotlib backends that
# map to the same GUI support
backend2gui["GTK"] = backend2gui["GTKCairo"] = "gtk"
backend2gui["WX"] = "wx"
backend2gui["CocoaAgg"] = "osx"


def do_enable_gui(guiname):
    from _pydev_bundle.pydev_versioncheck import versionok_for_gui

    if versionok_for_gui():
        try:
            from pydev_ipython.inputhook import enable_gui

            enable_gui(guiname)
        except:
            sys.stderr.write("Failed to enable GUI event loop integration for '%s'\n" % guiname)
            pydev_log.exception()
    elif guiname not in ["none", "", None]:
        # Only print a warning if the guiname was going to do something
        sys.stderr.write("Debug console: Python version does not support GUI event loop integration for '%s'\n" % guiname)
    # Return value does not matter, so return back what was sent
    return guiname


def find_gui_and_backend():
    """Return the gui and mpl backend."""
    matplotlib = sys.modules["matplotlib"]
    # WARNING: this assumes matplotlib 1.1 or newer!!
    backend = matplotlib.rcParams["backend"]

    # Translate to the real case as in 3.9 the case was forced to lowercase
    # but our internal mapping is in the original case.
    realcase_backend = lowercase_convert.get(backend, backend)

    # In this case, we need to find what the appropriate gui selection call
    # should be for IPython, so we can activate inputhook accordingly
    gui = backend2gui.get(realcase_backend, None)
    return gui, backend


def _get_major_version(module):
    return int(module.__version__.split('.')[0])


def _get_minor_version(module):
    return int(module.__version__.split('.')[1])


def is_interactive_backend(backend):
    """Check if backend is interactive"""
    matplotlib = sys.modules["matplotlib"]
    new_api_version = (3, 9)
    installed_version = (
        _get_major_version(matplotlib),
        _get_minor_version(matplotlib)
    )

    if installed_version >= new_api_version:
        interactive_bk = matplotlib.backends.backend_registry.list_builtin(
            matplotlib.backends.BackendFilter.INTERACTIVE)
        non_interactive_bk = matplotlib.backends.backend_registry.list_builtin(
            matplotlib.backends.BackendFilter.NON_INTERACTIVE)
    else:
        from matplotlib.rcsetup import interactive_bk, non_interactive_bk  # @UnresolvedImport

    if backend in interactive_bk:
        return True
    elif backend in non_interactive_bk:
        return False
    else:
        return matplotlib.is_interactive()


def patch_use(enable_gui_function):
    """Patch matplotlib function 'use'"""
    matplotlib = sys.modules["matplotlib"]

    def patched_use(*args, **kwargs):
        matplotlib.real_use(*args, **kwargs)
        gui, backend = find_gui_and_backend()
        enable_gui_function(gui)

    matplotlib.real_use = matplotlib.use
    matplotlib.use = patched_use


def patch_is_interactive():
    """Patch matplotlib function 'use'"""
    matplotlib = sys.modules["matplotlib"]

    def patched_is_interactive():
        return matplotlib.rcParams["interactive"]

    matplotlib.real_is_interactive = matplotlib.is_interactive
    matplotlib.is_interactive = patched_is_interactive


def activate_matplotlib(enable_gui_function):
    """Set interactive to True for interactive backends.
    enable_gui_function - Function which enables gui, should be run in the main thread.
    """
    matplotlib = sys.modules["matplotlib"]
    gui, backend = find_gui_and_backend()
    is_interactive = is_interactive_backend(backend)
    if is_interactive:
        enable_gui_function(gui)
        if not matplotlib.is_interactive():
            sys.stdout.write("Backend %s is interactive backend. Turning interactive mode on.\n" % backend)
        matplotlib.interactive(True)
    else:
        if matplotlib.is_interactive():
            sys.stdout.write("Backend %s is non-interactive backend. Turning interactive mode off.\n" % backend)
        matplotlib.interactive(False)
    patch_use(enable_gui_function)
    patch_is_interactive()


def flag_calls(func):
    """Wrap a function to detect and flag when it gets called.

    This is a decorator which takes a function and wraps it in a function with
    a 'called' attribute. wrapper.called is initialized to False.

    The wrapper.called attribute is set to False right before each call to the
    wrapped function, so if the call fails it remains False.  After the call
    completes, wrapper.called is set to True and the output is returned.

    Testing for truth in wrapper.called allows you to determine if a call to
    func() was attempted and succeeded."""

    # don't wrap twice
    if hasattr(func, "called"):
        return func

    def wrapper(*args, **kw):
        wrapper.called = False
        out = func(*args, **kw)
        wrapper.called = True
        return out

    wrapper.called = False
    wrapper.__doc__ = func.__doc__
    return wrapper


def activate_pylab():
    pylab = sys.modules["pylab"]
    pylab.show._needmain = False
    # We need to detect at runtime whether show() is called by the user.
    # For this, we wrap it into a decorator which adds a 'called' flag.
    pylab.draw_if_interactive = flag_calls(pylab.draw_if_interactive)


def activate_pyplot():
    pyplot = sys.modules["matplotlib.pyplot"]
    pyplot.show._needmain = False
    # We need to detect at runtime whether show() is called by the user.
    # For this, we wrap it into a decorator which adds a 'called' flag.
    pyplot.draw_if_interactive = flag_calls(pyplot.draw_if_interactive)
